# -*- coding: utf-8 -*-
import copy
from logging import getLogger

from tg.configuration import milestones
from tg.configuration.utils import TGConfigError
from tg.support.converters import asbool

from ..base import (
    BeforeConfigConfigurationAction,
    ConfigReadyConfigurationAction,
    ConfigurationComponent,
)

log = getLogger(__name__)

__all__ = ("TemplateRenderingConfigurationComponent",)


class TemplateRenderingConfigurationComponent(ConfigurationComponent):
    """Provides support for rendering engines.

    The available options are:

        - ``use_dotted_templatenames`` -> (``True``/``False``) Use template names as packages in
          @expose instead of file paths. This is usually the default unless TG is started
          in Minimal Mode.
        - ``auto_reload_templates`` -> (``True``/``False``) Automatically reload template files
          if they change. Should usually be disabled on production for performance reasons.
        - ``tg.strict_tmpl_context`` -> (``True``/``False``) Should ``tg.tmpl_context`` be
          strict and complain about missing value or should it always just return empty values
          for missing ones?
        - ``renderers`` -> (``list(str)``) List of template engines that should be enabled.
        - ``default_renderer`` -> (``str``) The default template engine to use when not explicitly
          specified by ``@expose`` decorations.

    Refer to each template engine renderer for specific configuration options.
    """

    id = "rendering"

    def __init__(self):
        super(TemplateRenderingConfigurationComponent, self).__init__()
        self.rendering_engines = {}
        self.rendering_engines_options = {}
        self.rendering_engines_without_vars = set()

    def get_coercion(self):
        return {
            "auto_reload_templates": asbool,
            "use_dotted_templatenames": asbool,
            "tg.strict_tmpl_context": asbool,
        }

    def get_defaults(self):
        return {
            "tg.strict_tmpl_context": True,
            "auto_reload_templates": True,
            "use_dotted_templatenames": False,
            "renderers": [],
            "default_renderer": "kajiki",
            "render_functions": {},
            "rendering_engines": {},
            "rendering_engines_without_vars": set(),
            "rendering_engines_options": {},
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_rendering),
            ConfigReadyConfigurationAction(self._setup_renderers),
        )

    def on_bind(self, configurator):
        from ...renderers.genshi import GenshiRenderer
        from ...renderers.jinja import JinjaRenderer
        from ...renderers.json import JSONRenderer
        from ...renderers.kajiki import KajikiRenderer
        from ...renderers.mako import MakoRenderer

        self.register_engine(JSONRenderer)
        self.register_engine(GenshiRenderer)
        self.register_engine(MakoRenderer)
        self.register_engine(JinjaRenderer)
        self.register_engine(KajikiRenderer)

    def register_engine(self, factory):
        """Registers a rendering engine ``factory``.

        Rendering engine factories are :class:`tg.renderers.base.RendererFactory`
        subclasses in charge of creating a rendering engine.

        """
        for engine, options in factory.engines.items():
            self.rendering_engines[engine] = factory
            self.rendering_engines_options[engine] = options
            if factory.with_tg_vars is False:
                self.rendering_engines_without_vars.add(engine)

    def _configure_rendering(self, conf, app):
        """Provides default configurations for renderers"""
        if "json" not in conf["renderers"]:
            conf["renderers"].append("json")

        if conf["default_renderer"] not in conf["renderers"]:
            first_renderer = conf["renderers"][0]
            log.warning(
                "Default renderer not in renders, "
                "automatically switching to %s" % first_renderer
            )
            conf["default_renderer"] = first_renderer

        conf["rendering_engines"] = copy.copy(self.rendering_engines)
        conf["rendering_engines_options"] = copy.copy(self.rendering_engines_options)
        conf["rendering_engines_without_vars"] = copy.copy(
            self.rendering_engines_without_vars
        )

    def _setup_renderers(self, conf, app):
        renderers = conf["renderers"]
        rendering_engines = conf["rendering_engines"]

        for renderer in renderers[:]:
            if renderer in rendering_engines:
                rendering_engine = rendering_engines[renderer]
                engines = rendering_engine.create(conf, conf["tg.app_globals"])
                if engines is None:
                    log.error(
                        "Failed to initialize %s template engine, removing it..."
                        % renderer
                    )
                    renderers.remove(renderer)
                else:
                    log.debug("Enabling renderer %s", renderer)
                    conf["render_functions"].update(engines)
            else:
                raise TGConfigError(
                    "This configuration object does "
                    "not support the %s renderer" % renderer
                )

        milestones.renderers_ready.reach()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

try:
    import sqlalchemy
    from sqlalchemy.engine import ResultProxy
except ImportError:  # pragma: no cover
    sqlalchemy = None
    ResultProxy = None

try:  # sa>=1.4
    from sqlalchemy.engine import Row

    try:
        from sqlalchemy.engine import LegacyRow as RowProxy
    except ImportError:  # pragma: no cover
        RowProxy = None
except ImportError:  # pragma: no cover
    Row = None
    try:  # sa<1.4
        from sqlalchemy.engine import RowProxy
    except ImportError:
        RowProxy = None


def is_saobject(obj):
    """Checks if the provided object is a SQLAlchemy model instance"""
    return sqlalchemy is not None and hasattr(obj, "_sa_class_manager")


def is_query_result(values):
    return ResultProxy is not None and isinstance(values, ResultProxy)


def is_query_row(obj):
    return (
        RowProxy is not None
        and isinstance(obj, RowProxy)
        or Row is not None
        and isinstance(obj, Row)
    )


def dictify(obj):
    """Converts a SQLAlchemy model instance to a dictionary"""
    if sqlalchemy is None:  # pragma: no cover
        raise RuntimeError("SQLAlchemy not available")

    if sqlalchemy.inspect(obj).detached:
        raise ValueError(
            "SQLAlchemy instance '%r' must be attached to a session." % obj
        )

    props = {}
    for key in obj.__dict__:
        if not key.startswith("_sa_"):
            props[key] = getattr(obj, key)
    return props



# -*- coding: utf-8 -*-
from logging import getLogger

from tg.support.converters import asbool
from tg.support.registry import RegistryManager

from ..base import AppReadyConfigurationAction, ConfigurationComponent

log = getLogger(__name__)

__all__ = ("RegistryConfigurationComponent",)


class RegistryConfigurationComponent(ConfigurationComponent):
    """Configure the request local context registry.

    This configures support for setting and restoring a clean
    turbogears context on each request. This makes so that
    ``tg.request``, ``tg.response`` and so on always refer to
    the data for current request.

    Options:

        * ``registry_streaming``: Enable streaming responses, thus restoring
                                  the registry at the end of the stream instead of
                                  as soon as the controller action returned.
                                  This is enabled by default.
        * ``debug``: Ensures that the registry is not discarded in case of an
                     exception. So that after the exception is possible to inspect
                     the state of the request that caused the exception.

    """

    id = "registry"

    def get_coercion(self):
        return {"registry_streaming": asbool, "debug": asbool}

    def get_defaults(self):
        return {
            "registry_streaming": True,
        }

    def get_actions(self):
        return (AppReadyConfigurationAction(self._add_registry_middleware),)

    def _add_registry_middleware(self, conf, app):
        # Establish the registry for this application
        return RegistryManager(
            app,
            streaming=conf.get("registry_streaming", True),
            preserve_exceptions=conf.get("debug", False),
        )



import logging
import sys

from ..configuration.utils import coerce_config
from ..request_local import Response
from ..support.converters import asbool, asint, aslist
from .base import ApplicationWrapper

log = logging.getLogger(__name__)


class ErrorPageApplicationWrapper(ApplicationWrapper):
    """Given an Application it intercepts the response code and shows a custom page.

    Supported options are:

        - ``errorpage.enabled``: Whenever the custom error page is enabled or not.
        - ``errorpage.status_codes``: List of HTTP errors that should be trapped.
          By default 403, 404, 500.
        - ``errorpage.content_types``: List of Content-Types for which the custom error
          page should be displayed. IE: ``["text/html"]``. An empty list means all.
          An entry with value ``None`` means no content type provided.
          Default is ``["text/html", None]``.
        - ``errorpage.handle_exceptions``: Whenever exceptions should be trapped and
          treated as a 500 error or not. By default this is ``True`` when ``debug=false``.
        - ``errorapge.path``: Path of the controller should be displayed in case of
          errors. By default ``/error/document``.
    """

    def __init__(self, handler, config):
        super(ErrorPageApplicationWrapper, self).__init__(handler, config)

        options = {
            "enabled": False,
            "status_codes": tuple(),
            "handle_exceptions": not asbool(config.get("debug", False)),
            "path": "/error/document",
            "content_types": ["text/html", None],
        }
        options.update(
            coerce_config(
                config,
                "errorpage.",
                {
                    "enabled": asbool,
                    "status_codes": aslist,
                    "handle_exceptions": asbool,
                    "content_types": aslist,
                },
            )
        )

        self.handle_error_enabled = options["enabled"]
        self.handle_status_codes = set(asint(s) for s in options["status_codes"])
        self.handle_exceptions = options["handle_exceptions"]
        self.handle_error_path = options["path"]
        self.handle_content_types = options["content_types"]

        if self.handle_exceptions and 500 not in self.handle_status_codes:
            self.handle_status_codes.add(500)

        log.debug(
            "ErrorPageApplicationWrapper enabled: %s -> %s",
            self.handle_error_enabled,
            options,
        )

    @property
    def injected(self):
        return self.handle_error_enabled

    def __call__(self, controller, environ, context):
        try:
            resp = self.next_handler(controller, environ, context)
        except Exception:
            if self.handle_exceptions is False:
                raise
            # Provide crash details to backlash
            environ["backlash.exc_environ"] = environ.copy()
            environ["backlash.exc_info"] = sys.exc_info()
            # Force response to a 500 Error, otherwise it will be a 200
            resp = context.response
            resp.status_code = 500

        if not environ.get("tg.status_code_redirect", True):
            # status_code_redirect disabled per this request
            return resp

        status_code = resp.status_code
        content_type = resp.content_type
        log.debug(
            "ErrorPageApplicationWrapper response: %s -> %s @ %s",
            environ["PATH_INFO"],
            status_code,
            content_type,
        )
        if status_code in self.handle_status_codes and (
            not self.handle_content_types or content_type in self.handle_content_types
        ):
            environ["tg.original_request"] = context.request.copy()
            environ["tg.original_response"] = resp

            # Reset the response, so the error controller starts
            # with a clean one and can provide the wished response.
            # The original one will be available in environ as
            # tg.original_response
            resp_options = context.config.get(
                "tg.response_options", Response._DEFAULT_RESPONSE_OPTIONS
            )
            context.response = Response(
                content_type=resp_options["content_type"],
                charset=resp_options["charset"],
                headers=resp_options["headers"],
                status_code=status_code,
            )

            environ["PATH_INFO"] = self.handle_error_path
            log.debug(
                "ErrorPageApplicationWrapper serving %s:%s",
                controller,
                self.handle_error_path,
            )
            resp = self.next_handler(controller, environ, context)

        return resp



import mimetypes
from datetime import datetime
from email.utils import mktime_tz, parsedate_tz
from os.path import getmtime, getsize, isfile, join, normcase, normpath
from time import gmtime, time

from repoze.lru import LRUCache
from webob.exc import HTTPBadRequest, HTTPForbidden, HTTPNotFound

_BLOCK_SIZE = 4096 * 64  # 256K

mimetypes.init()


class _FileIter(object):
    def __init__(self, file, block_size):
        self.file = file
        self.block_size = block_size

    def __iter__(self):
        return self

    def next(self):
        val = self.file.read(self.block_size)
        if not val:
            raise StopIteration
        return val

    __next__ = next  # py3

    def close(self):
        self.file.close()


class FileServeApp(object):
    """
    Serves a static filelike object.
    """

    def __init__(self, path, cache_max_age):
        self.path = path

        try:
            self.last_modified = getmtime(path)
            self.content_length = getsize(path)
        except (IOError, OSError):
            self.path = None

        if self.path is not None:
            content_type, content_encoding = mimetypes.guess_type(path, strict=False)
            if content_type is None:
                content_type = "application/octet-stream"

            self.content_type = content_type
            self.content_encoding = content_encoding

        if cache_max_age is not None:
            self.cache_expires = cache_max_age

    def generate_etag(self):
        return '"%s-%s"' % (self.last_modified, self.content_length)

    def parse_date(self, value):
        try:
            return mktime_tz(parsedate_tz(value))
        except (TypeError, OverflowError):
            raise HTTPBadRequest(
                ("Received an ill-formed timestamp for %s: %s\r\n") % (self.path, value)
            )

    @classmethod
    def make_date(cls, d):
        if isinstance(d, datetime):
            d = d.utctimetuple()
        else:
            d = gmtime(d)

        return "%s, %02d%s%s%s%s %02d:%02d:%02d GMT" % (
            ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")[d.tm_wday],
            d.tm_mday,
            " ",
            (
                "Jan",
                "Feb",
                "Mar",
                "Apr",
                "May",
                "Jun",
                "Jul",
                "Aug",
                "Sep",
                "Oct",
                "Nov",
                "Dec",
            )[d.tm_mon - 1],
            " ",
            str(d.tm_year),
            d.tm_hour,
            d.tm_min,
            d.tm_sec,
        )

    def has_been_modified(self, environ, etag, last_modified):
        if environ["REQUEST_METHOD"] not in ("GET", "HEAD"):
            return False

        unmodified = False

        modified_since = environ.get("HTTP_IF_MODIFIED_SINCE")
        if modified_since:
            modified_since = self.parse_date(modified_since)
            if last_modified and last_modified <= modified_since:
                unmodified = True

        if_none_match = environ.get("HTTP_IF_NONE_MATCH")
        if if_none_match and etag == if_none_match:
            unmodified = True

        return not unmodified

    def __call__(self, environ, start_response):
        try:
            file = open(self.path, "rb")
        except (IOError, OSError, TypeError) as e:
            return HTTPForbidden("You are not permitted to view this file (%s)" % e)(
                environ, start_response
            )

        headers = []
        timeout = self.cache_expires
        etag = self.generate_etag()
        headers += [
            ("Etag", "%s" % etag),
            ("Cache-Control", "max-age=%d, public" % timeout),
        ]

        if not self.has_been_modified(environ, etag, self.last_modified):
            file.close()
            start_response("304 Not Modified", headers)
            return []

        headers.extend(
            (
                ("Expires", self.make_date(time() + timeout)),
                ("Content-Type", self.content_type),
                ("Content-Length", str(self.content_length)),
                ("Last-Modified", self.make_date(self.last_modified)),
            )
        )
        start_response("200 OK", headers)
        return environ.get("wsgi.file_wrapper", _FileIter)(file, _BLOCK_SIZE)


INVALID_PATH_PARTS = set(["..", "."]).intersection


class StaticsMiddleware(object):
    def _adapt_path(self, path):
        return normcase(normpath(path))

    def __init__(self, app, root_dir, cache_max_age=3600):
        self.app = app
        self.cache_max_age = cache_max_age
        self.doc_root = self._adapt_path(root_dir)
        self.paths_cache = LRUCache(1024)

    def __call__(self, environ, start_response):
        full_path = environ["PATH_INFO"]
        filepath = self.paths_cache.get(full_path)

        if filepath is None:
            path = full_path.split("/")
            if INVALID_PATH_PARTS(path):
                return HTTPNotFound("Out of bounds: %s" % environ["PATH_INFO"])(
                    environ, start_response
                )
            filepath = self._adapt_path(join(self.doc_root, *path))
            self.paths_cache.put(full_path, filepath)

        if isfile(filepath):
            return FileServeApp(filepath, self.cache_max_age)(environ, start_response)

        return self.app(environ, start_response)



from ..caching import cached_property
from .decoration import Decoration
from .decorators import (
    before_call,
    before_render,
    before_validate,
    cached,
    decode_params,
    expose,
    https,
    override_template,
    paginate,
    require,
    use_custom_format,
    validate,
    with_engine,
    with_trailing_slash,
    without_trailing_slash,
)

__all__ = (
    "Decoration",
    "before_call",
    "before_render",
    "before_validate",
    "cached",
    "decode_params",
    "expose",
    "override_template",
    "paginate",
    "https",
    "require",
    "use_custom_format",
    "validate",
    "with_engine",
    "with_trailing_slash",
    "without_trailing_slash",
    "cached_property",
)



def get_partial_dict(
    prefix, dictionary, container_type=dict, ignore_missing=False, pop_keys=False
):
    """Given a dictionary and a prefix, return a Bunch, with just items
    that start with prefix

    The returned dictionary will have 'prefix.' stripped so::

        get_partial_dict('prefix', {'prefix.xyz':1, 'prefix.zyx':2, 'xy':3})

    would return::

        {'xyz':1,'zyx':2}
    """

    match = prefix + "."
    n = len(match)

    new_dict = container_type(
        ((key[n:], dictionary[key]) for key in dictionary if key.startswith(match))
    )

    if pop_keys:
        for key in list(dictionary.keys()):
            if key.startswith(match):
                dictionary.pop(key, None)

    if new_dict:
        return new_dict
    else:
        if ignore_missing:
            return {}
        raise AttributeError(prefix)


class Bunch(dict):
    """A dictionary that provides attribute-style access."""

    def __getitem__(self, key):
        return dict.__getitem__(self, key)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            return get_partial_dict(name, self, Bunch)

    __setattr__ = dict.__setitem__

    def __delattr__(self, name):
        try:
            del self[name]
        except KeyError:
            raise AttributeError(name)



# -*- coding: utf-8 -*-
"""
Utilities for TurboGears hooks management.

Provides a consistent API to register and execute hooks.

"""

import atexit
from logging import getLogger

from ..configuration.milestones import config_ready, renderers_ready
from ..configuration.utils import TGConfigError
from ..decorators import Decoration
from ..util.instance_method import default_im_func

log = getLogger(__name__)


class HooksNamespace(object):
    """Manages hooks registrations and notifications"""

    def __init__(self):
        self._hooks = dict()
        atexit.register(self._atexit)

    def _clear(self):
        self._hooks.clear()

    def _atexit(self):
        for func in self._hooks.get("shutdown", tuple()):
            func()

    def _call_handler(self, hook_name, trap_exceptions, func, args, kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            if trap_exceptions is True:
                log.exception(
                    "Trapped Exception while handling %s -> %s", hook_name, func
                )
            else:
                raise

    def register(self, hook_name, func, controller=None):
        """Registers a TurboGears hook.

        Given an hook name and a function it registers the provided
        function for that role. For a complete list of hooks
        provided by default have a look at :ref:`hooks_and_events`.

        It permits to register hooks both application wide
        or for specific controllers::

            tg.hooks.register('before_render', hook_func, controller=RootController.index)
            tg.hooks.register('startup', startup_function)

        """
        if controller is not None and hook_name in (
            "startup",
            "shutdown",
            "initialized_config",
            "before_wsgi_middlewares",
            "after_wsgi_middlewares",
        ):
            raise TGConfigError(
                "Startup and Shutdown hooks cannot be registered on controllers"
            )

        if hook_name == "controller_wrapper":
            raise ValueError(
                "dispatch component must be used to register controller wrappers"
            )

        if controller is None:
            config_ready.register(_ApplicationHookRegistration(self, hook_name, func))
        else:
            controller = default_im_func(controller)
            renderers_ready.register(
                _ControllerHookRegistration(controller, hook_name, func)
            )

    def disconnect(self, hook_name, func, controller=None):
        """Disconnect an hook.

        The registered function is removed from the hook notification list.
        """
        if controller is None:
            registrations = self._hooks.get(hook_name, [])
        else:
            deco = Decoration.get_decoration(controller)
            registrations = deco.hooks.get(hook_name, [])

        try:
            registrations.remove(func)
        except ValueError:
            pass

    def notify(
        self, hook_name, args=None, kwargs=None, controller=None, trap_exceptions=False
    ):
        """Notifies a TurboGears hook.

        Each function registered for the given hook will be executed,
        ``args`` and ``kwargs`` will be passed to the registered functions
        as arguments.

        It permits to notify both application hooks::

            tg.hooks.notify('custom_global_hook')

        Or controller hooks::

            tg.hooks.notify('before_render', args=(remainder, params, output),
                            controller=RootController.index)

        """
        args = args or []
        kwargs = kwargs or {}

        try:
            syswide_hooks = self._hooks[hook_name]
        except KeyError:  # pragma: no cover
            pass
        else:
            for func in syswide_hooks:
                self._call_handler(hook_name, trap_exceptions, func, args, kwargs)

        if controller is not None:
            controller = default_im_func(controller)
            deco = Decoration.get_decoration(controller)
            for func in deco.hooks.get(hook_name, []):
                self._call_handler(hook_name, trap_exceptions, func, args, kwargs)

    def notify_with_value(self, hook_name, value, controller=None):
        """Notifies a TurboGears hook which is expected to return a value.

        hooks with values are expected to accept an input value an return
        a replacement for it. Each registered function will receive as input
        the value returned by the previous function in chain.

        The resulting value will be returned by the ``notify_with_value``
        call itself::

            app = tg.hooks.notify_with_value('before_config', app)

        """
        try:
            syswide_hooks = self._hooks[hook_name]
        except KeyError:  # pragma: no cover
            pass
        else:
            for func in syswide_hooks:
                value = func(value)

        if controller is not None:
            controller = default_im_func(controller)
            deco = Decoration.get_decoration(controller)
            for func in deco.hooks[hook_name]:
                value = func(value)

        return value


class _ApplicationHookRegistration(object):
    def __init__(self, hooks_namespace, hook_name, func):
        self.hook_name = hook_name
        self.func = func
        self.hooks_namespace = hooks_namespace

    def __repr__(self):
        return "<ApplicationHookRegistration: %r %r>" % (self.hook_name, self.func)

    def __call__(self):
        log.debug(
            "Registering %s for application wide hook %s", self.func, self.hook_name
        )
        hooks = self.hooks_namespace._hooks
        hooks.setdefault(self.hook_name, []).append(self.func)


class _ControllerHookRegistration(object):
    def __init__(self, controller, hook_name, func):
        self.controller = controller
        self.hook_name = hook_name
        self.func = func

    def __repr__(self):
        return "<ControllerHookRegistration: %r %r>" % (self.hook_name, self.func)

    def __call__(self):
        log.debug(
            "Registering %s for hook %s on controller %s",
            self.func,
            self.hook_name,
            self.controller,
        )
        deco = Decoration.get_decoration(self.controller)
        deco._register_hook(self.hook_name, self.func)


class _TGGlobalHooksNamespace(HooksNamespace):
    pass


hooks = _TGGlobalHooksNamespace()



import base64
import binascii
import hmac
import pickle
import urllib.parse
from hashlib import sha1

from webob import Request as WebObRequest
from webob import Response as WebObResponse
from webob.request import PATH_SAFE

from .caching import cached_property
from .support.objectproxy import TurboGearsObjectProxy
from .support.registry import DispatchingConfig, StackedObjectProxy


class Request(WebObRequest):
    """WebOb Request subclass

    The WebOb :class:`webob.Request` has no charset, or other defaults. This subclass
    adds defaults, along with several methods for backwards
    compatibility with paste.wsgiwrappers.WSGIRequest.

    """

    def _fast_setattr(self, name, value):
        object.__setattr__(self, name, value)

    def languages_best_match(self, fallback=None):
        al = self.accept_language
        try:
            items = [i for i, q in sorted(al._parsed, key=lambda iq: -iq[1])]
        except (AttributeError, TypeError):
            # NilAccept has no _parsed, here for test units
            items = []

        if fallback:
            items.append(fallback)

        return items

    @cached_property
    def dispatch_state(self):
        """Details and info about dispatcher that handled this request."""
        return self._dispatch_state

    @cached_property
    def controller_url(self):
        """Url of the current controller."""
        state = self._dispatch_state
        return "/".join(state.path[: -len(state.remainder)])

    @cached_property
    def plain_languages(self):
        """Return the list of browser preferred languages"""
        return self.languages_best_match()

    @cached_property
    def languages(self):
        """Return the list of browser preferred languages ensuring that ``i18n.lang`` is listed."""
        return self.languages_best_match(self._language)

    @property
    def response_type(self):
        """Expected response content type when URL Extensions are enabled.

        In case URL Request Extension is enabled this will be the content type
        of the expected response. ``disable_request_extensions`` drives
        this is enabled or not.
        """
        return self._response_type

    @property
    def response_ext(self):
        """URL extension when URL Extensions are enabled.

        In case URL Request Extension is enabled this will be the extension of the url.
        ``disable_request_extensions`` drives this is enabled or not.
        """
        return self._response_ext

    def match_accept(self, mimetypes):
        return self.accept.best_match(mimetypes)

    def signed_cookie(self, name, secret):
        """Extract a signed cookie of ``name`` from the request

        The cookie is expected to have been created with
        ``Response.signed_cookie``, and the ``secret`` should be the
        same as the one used to sign it.

        Any failure in the signature of the data will result in None
        being returned.

        """
        cookie = self.cookies.get(name)
        if not cookie:
            return

        secret = secret.encode("ascii")
        try:
            # Keep base64.decodestring for backward compatibility.
            # In the past Response.signed_cookie used base64.encodestring which could
            # lead to multiple lines in the encoded text. That's in theory unsupported
            # as cookies can't have newlines, but we still want to be able to decode it.
            try:
                base64decode = base64.decodebytes
            except AttributeError:  # pragma: no cover
                base64decode = base64.decodestring
            sig, pickled = cookie[:40], base64decode(cookie[40:].encode("ascii"))
        except binascii.Error:  # pragma: no cover
            # Badly formed data can make base64 die
            return

        if hmac.new(secret, pickled, sha1).hexdigest() == sig:
            return pickle.loads(pickled)

    @cached_property
    def args_params(self):
        """Arguments used for dispatching the request.

        This mixes GET and POST arguments.
        """
        # This was: dict(((str(n), v) for n,v in self.params.mixed().items()))
        # so that keys were all strings making possible to use them as arguments.
        # Now it seems that all keys are always strings, did WebOb change behavior?
        return self.params.mixed()

    @property
    def quoted_path_info(self):
        """PATH used for dispatching the request."""
        bpath = str(self.path_info).encode(self.url_encoding)
        return urllib.parse.quote(bpath, PATH_SAFE)

    def disable_error_pages(self):
        """Disable custom error pages for the current request.

        This will forward your response as is bypassing the :class:`.ErrorPageApplicationWrapper`
        """
        self.environ["tg.status_code_redirect"] = False

    def disable_auth_challenger(self):
        """Disable authentication challenger for current request.

        This will forward your response as is in case of 401 bypassing any
        repoze.who challenger.
        """
        self.environ["tg.skip_auth_challenge"] = True


class Response(WebObResponse):
    """WebOb Response subclass"""

    _DEFAULT_RESPONSE_OPTIONS = dict(
        content_type="text/html",
        charset="utf-8",
        headers={
            "Cache-Control": "no-cache",
            "Pragma": "no-cache",
            "Content-Type": None,
            "Content-Length": "0",
        },
    )

    content = WebObResponse.body

    def wsgi_response(self):
        return self.status, self.headers, self.body

    def signed_cookie(self, name, data, secret, **kwargs):
        """Save a signed cookie with ``secret`` signature

        Saves a signed cookie of the pickled data. All other keyword
        arguments that ``WebOb.set_cookie`` accepts are usable and
        passed to the WebOb set_cookie method after creating the signed
        cookie value.

        """
        secret = secret.encode("ascii")

        pickled = pickle.dumps(data, pickle.HIGHEST_PROTOCOL)
        sig = hmac.new(secret, pickled, sha1).hexdigest().encode("ascii")
        cookie_value = sig + base64.b64encode(pickled)
        self.set_cookie(name, cookie_value, **kwargs)

    @property
    def content_type(self):
        return WebObResponse.content_type.__get__(self, type(self))

    @content_type.setter
    def content_type(self, value):
        WebObResponse.content_type.__set__(self, value)

    @content_type.deleter
    def content_type(self):
        WebObResponse.content_type.__delete__(self)


config = DispatchingConfig()
context = StackedObjectProxy(name="context")


class TurboGearsContextMember(TurboGearsObjectProxy):
    """Member of the TurboGears request context.

    Provides access to turbogears context members
    like request, response, template context and so on

    """

    def __init__(self, name):
        self.__dict__["name"] = name

    def _current_obj(self):
        return getattr(context, self.name)


request = TurboGearsContextMember(name="request")
app_globals = TurboGearsContextMember(name="app_globals")
cache = TurboGearsContextMember(name="cache")
response = TurboGearsContextMember(name="response")
session = TurboGearsContextMember(name="session")
tmpl_context = TurboGearsContextMember(name="tmpl_context")
url = TurboGearsContextMember(name="url")
translator = TurboGearsContextMember(name="translator")

__all__ = [
    "app_globals",
    "request",
    "response",
    "tmpl_context",
    "session",
    "cache",
    "translator",
    "url",
    "config",
]



"""Configuration Helpers for TurboGears 2"""

import logging
import warnings

log = logging.getLogger(__name__)


class AppConfig(object):
    """Backward compatible Application Configurator.

    This allows to configure a TurboGears2 application in a way
    that is compatible with existing >=2.1,<=2.3 applications.

    All the attributes and keys you will set into the AppConfig
    will be used to build the blueprint used by a
    :class:`.FullStackApplicationConfigurator`
    to configure a new :class:`.TGApp`.

    .. deprecated:: 2.4.0
        Use :class:`.FullStackApplicationConfigurator` instead.
    """

    __slots__ = ("_configurator",)

    # Attributes and properties that are automatically returned as a view
    # This mostly handles backward compatibility with some oddities of
    # TG<2.4 where some config properties where flat and some were subdicts.
    VIEWS_ATTRIBUTES = set(("sa_auth",))

    def __init__(self, **kwargs):
        from ..configurator import FullStackApplicationConfigurator
        from ..support.hooks import hooks as tg_hooks

        self._configurator = FullStackApplicationConfigurator()

        if kwargs.pop("minimal", False):
            self._configurator.update_blueprint(
                {
                    "i18n.enabled": False,
                    "session.enabled": False,
                    "sa_auth.enabled": False,
                    "cache.enabled": False,
                    "tw2.enabled": False,
                    "use_ming": False,
                    "use_sqlalchemy": False,
                    "tm.enabled": False,
                    "errorpage.enabled": False,
                    "make_body_seekable": False,
                    "trace_slowreqs.enable": False,
                    "trace_errors.enable": False,
                    "serve_static": False,
                    "debug": False,
                }
            )
        self._configurator.update_blueprint(kwargs)

        def _on_config_ready(_, conf):
            self.after_init_config(conf)

        tg_hooks.register("initialized_config", _on_config_ready)

        def _startup_hook(*args, **kwargs):
            tg_hooks.notify("startup", trap_exceptions=True)

        tg_hooks.register("initialized_config", _startup_hook)

        def _before_config_hook(app):
            return tg_hooks.notify_with_value("before_config", app)

        tg_hooks.register("before_wsgi_middlewares", _before_config_hook)

        def _after_config_hook(app):
            return tg_hooks.notify_with_value("after_config", app)

        tg_hooks.register("after_wsgi_middlewares", _after_config_hook)

    def after_init_config(self, conf):
        """
        Override this method to set up configuration variables at the application
        level.  This method will be called after your configuration object has
        been initialized on startup.  Here is how you would use it to override
        the default setting of tg.strict_tmpl_context ::

            from tg import Configurator

            class MyAppConfigurator(Configurator):
                def after_init_config(self, conf):
                    conf['tg.strict_tmpl_context'] = False

            base_config = MyAppConfig()

        """
        pass

    def __setitem__(self, key, value):
        if key in self.VIEWS_ATTRIBUTES:
            self.get_view(key).update(value)
        else:
            self._configurator.update_blueprint({key: value})

    def __getitem__(self, item):
        if item in self.VIEWS_ATTRIBUTES:
            return self.get_view(item)
        return self._configurator.get_blueprint_value(item)

    def get(self, k, d=None):
        try:
            return self[k]
        except KeyError:
            return d

    def get_view(self, item):
        return self._configurator.get_blueprint_view(item)

    def __setattr__(self, key, value):
        if key not in self.__slots__:
            self.__setitem__(key, value)
        else:
            object.__setattr__(self, key, value)
        return value

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(item)

    def register_hook(self, hookname, handler, controller=None):
        from ..support.hooks import hooks as tg_hooks

        tg_hooks.register(hookname, handler, controller=controller)

    def register_wrapper(self, wrapper, after=None):
        self._configurator.register_application_wrapper(wrapper, after)

    def register_rendering_engine(self, factory):
        self._configurator.get_component("rendering").register_engine(factory)

    def register_controller_wrapper(self, wrapper, controller=None):
        self._configurator.get_component("dispatch").register_controller_wrapper(
            wrapper, controller
        )

    def make_load_environment(self):
        """Return a load_environment function.

        The returned load_environment function can be called to configure
        the TurboGears runtime environment for this particular application.
        You can do this dynamically with multiple nested TG applications
        if necessary.

        """

        def _load_environment(global_conf, app_conf):
            conf = self._configurator.configure(global_conf, app_conf)
            self._configurator.setup(conf)
            return conf

        return _load_environment

    def setup_tg_wsgi_app(self, load_environment=None):
        """Creates a TGApp Factory, with the required load_environment.

        ``load_environment``
            A callable, which sets up the basic evironment
            needed for the application. A default environment is configured otherwise.

        The returned factory function accepts:

            * ``global_conf``: Dictionary with options that should be added to configuration.
            * ``wrap_app``: A function that can wrap application and return a new WSGI app.
            * ``**app_conf``: Keyword arguments that will be passed as configuration options.

        """
        warnings.warn(
            "Using AppConfig to create apps is deprecated in favor of "
            "tg.FullStackApplicationConfigurator and will be removed.",
            DeprecationWarning,
        )

        def make_base_app(global_conf=None, wrap_app=None, **app_conf):
            # Configure the Application environment
            init_config = load_environment
            if init_config is None:
                init_config = self.make_load_environment()
            return self._configurator._make_app(
                init_config(global_conf or {}, app_conf), wrap_app
            )

        return make_base_app

    def make_wsgi_app(self, **kwargs):
        """Creates a new TGApp.

        Only accepted argument is ``wrap_app`` as a keyword argument,
        that can be a callable used to wrap the TGApp in middlewares and
        return a new WSGI application.

        All remaining ``kwargs`` will be added as configuration options to the
        application in addition to those specified in ``AppConfig`` itself.

        """
        warnings.warn(
            "Using AppConfig to create apps is deprecated in favor of "
            "tg.FullStackApplicationConfigurator and will be removed.",
            DeprecationWarning,
        )

        # wrap_app is an argument to make_wsgi_app, not a configuration option.
        wrap_app = kwargs.pop("wrap_app", None)
        return self._configurator.make_wsgi_app({}, kwargs, wrap_app=wrap_app)



import json
from decimal import Decimal

import pytest
from webob.multidict import MultiDict

from tests.base import utcnow
from tg import jsonify, lurl
from tg.util import LazyString
from tg.util.webtest import test_context


class Foo(object):
    def __init__(self, bar):
        self.bar = bar

class Bar(object):
    def __init__(self, bar):
        self.bar = bar
    def __json__(self):
        return 'bar-%s' % self.bar

class Baz(object):
    pass

def test_string():
    d = "string"
    encoded = jsonify.encode(d)
    assert encoded == '"string"'

def test_list():
    d = ['a', 1, 'b', 2]

    with pytest.raises(jsonify.JsonEncodeError):
        encoded = jsonify.encode(d)
    #assert encoded == '["a", 1, "b", 2]'

def test_list_iter():
    d = list(range(3))

    with pytest.raises(jsonify.JsonEncodeError):
        encoded = jsonify.encode_iter(d)
    #assert ''.join(jsonify.encode_iter(d)) == jsonify.encode(d)

def test_list_allowed_iter():
    lists_encoder = jsonify.JSONEncoder(allow_lists=True)
    d = list(range(3))
    encoded = jsonify.encode_iter(d, lists_encoder)
    assert ''.join(encoded) == '[0, 1, 2]'

def test_dictionary():
    d = {'a': 1, 'b': 2}
    encoded = jsonify.encode(d)
    expected = json.dumps(json.loads('{"a": 1, "b": 2}'))
    assert encoded == expected


def test_nospecificjson():
    b = Baz()

    with pytest.raises(jsonify.JsonEncodeError):
        encoded = jsonify.encode(b)

def test_exlicitjson():
    b = Bar("bq")
    encoded = jsonify.encode(b)
    assert encoded == '"bar-bq"'

def test_exlicitjson_in_list():
    b = Bar("bq")
    d = [b]
    with pytest.raises(jsonify.JsonEncodeError):
        encoded = jsonify.encode(d)

def test_exlicitjson_in_dict():
    b = Bar("bq")
    d = {"b": b}
    encoded = jsonify.encode(d)
    assert encoded == '{"b": "bar-bq"}'

def test_datetime():
    d = utcnow()
    encoded = jsonify.encode({'date':d})
    assert str(d.year) in encoded, (str(d), encoded)

def test_datetime_iso():
    isodates_encoder = jsonify.JSONEncoder(isodates=True)

    d = utcnow()
    encoded = jsonify.encode({'date': d}, encoder=isodates_encoder)

    isoformat_without_millis = json.dumps({'date': d.isoformat()[:19]})
    assert isoformat_without_millis == encoded, (isoformat_without_millis, encoded)
    assert 'T' in encoded, encoded

def test_date_iso():
    isodates_encoder = jsonify.JSONEncoder(isodates=True)

    d = utcnow().date()
    encoded = jsonify.encode({'date': d}, encoder=isodates_encoder)

    isoformat_without_millis = json.dumps({'date': d.isoformat()})
    assert isoformat_without_millis == encoded, (isoformat_without_millis, encoded)

    loaded_date = json.loads(encoded)
    assert len(loaded_date['date'].split('-')) == 3

def test_datetime_time():
    d = utcnow().time()
    encoded = jsonify.encode({'date':d})
    assert str(d.hour) in encoded, (str(d), encoded)

def test_datetime_time_iso():
    isodates_encoder = jsonify.JSONEncoder(isodates=True)

    d = utcnow().time()
    encoded = jsonify.encode({'date': d}, encoder=isodates_encoder)

    isoformat_without_millis = json.dumps({'date': d.isoformat()[:8]})
    assert isoformat_without_millis == encoded, (isoformat_without_millis, encoded)

def test_decimal():
    d = Decimal('3.14')
    encoded = jsonify.encode({'dec':d})
    assert '3.14' in encoded

def test_objectid():
    try:
        from bson import ObjectId
    except ImportError:
        raise pytest.skip("MongoDB not supported on this system")

    d = ObjectId('507f1f77bcf86cd799439011')
    encoded = jsonify.encode({'oid':d})
    assert encoded == '{"oid": "%s"}' % d, encoded

def test_multidict():
    d = MultiDict({'v':1})
    encoded = jsonify.encode({'md':d})
    assert encoded == '{"md": {"v": 1}}', encoded

def test_json_encode_lazy_url():
    with test_context(None, '/'):
        url = lurl('/test')
        encoded = jsonify.encode({'url': url})
        assert encoded == '{"url": "/test"}', encoded

def test_json_encode_lazy_string():
    text = LazyString(lambda: 'TEST_STRING')
    encoded = jsonify.encode({'text': text})
    assert encoded == '{"text": "TEST_STRING"}', encoded

def test_json_encode_generators():
    encoded = jsonify.encode({'values': (v for v in [1, 2, 3])})
    assert encoded == '{"values": [1, 2, 3]}', encoded



import logging

from ..request_local import Request
from .statics import StaticsMiddleware

log = logging.getLogger(__name__)


class SeekableRequestBodyMiddleware(object):
    def __init__(self, app):
        self.app = app

    def _stream_response(self, data):
        try:
            for chunk in data:
                yield chunk
        finally:
            if hasattr(data, "close"):  # pragma: no cover
                data.close()

    def __call__(self, environ, start_response):
        log.debug("Making request body seekable")
        Request(environ).make_body_seekable()
        return self._stream_response(self.app(environ, start_response))


class DBSessionRemoverMiddleware(object):
    def __init__(self, DBSession, app):
        self.app = app
        self.DBSession = DBSession

    def _stream_response(self, data):
        try:
            for chunk in data:
                yield chunk
        finally:
            log.debug("Removing DBSession from current thread")
            if hasattr(data, "close"):
                data.close()
            self.DBSession.remove()

    def __call__(self, environ, start_response):
        try:
            return self._stream_response(self.app(environ, start_response))
        except Exception:
            log.debug("Removing DBSession from current thread")
            self.DBSession.remove()
            raise


class MingSessionRemoverMiddleware(object):
    def __init__(self, ThreadLocalODMSession, app):
        self.app = app
        self.ThreadLocalODMSession = ThreadLocalODMSession

    def _stream_response(self, data):
        try:
            for chunk in data:
                yield chunk
        finally:
            log.debug("Removing ThreadLocalODMSession from current thread")
            if hasattr(data, "close"):
                data.close()
            self.ThreadLocalODMSession.close_all()

    def __call__(self, environ, start_response):
        try:
            return self._stream_response(self.app(environ, start_response))
        except Exception:
            log.debug("Removing ThreadLocalODMSession from current thread")
            self.ThreadLocalODMSession.close_all()
            raise


__all__ = [
    "StaticsMiddleware",
    "SeekableRequestBodyMiddleware",
    "DBSessionRemoverMiddleware",
    "MingSessionRemoverMiddleware",
]



import re

from webob.exc import status_map

from ..exceptions import HTTPFound
from ..jsonify import encode as json_encode
from ..request_local import Response
from ..request_local import request as tg_request
from ..request_local import response as tg_response
from ..support.url import url


def redirect(base_url="/", params=None, redirect_with=HTTPFound, scheme=None):
    """Generate an HTTP redirect.

    The function raises an exception internally,
    which is handled by the framework. The URL may be either absolute (e.g.
    http://example.com or /myfile.html) or relative. Relative URLs are
    automatically converted to absolute URLs. Parameters may be specified,
    which are appended to the URL. This causes an external redirect via the
    browser; if the request is POST, the browser will issue GET for the
    second request.
    """

    if params is None:
        params = {}

    new_url = url(base_url, params=params, scheme=scheme)
    raise redirect_with(location=new_url)


IF_NONE_MATCH = re.compile(r'(?:W/)?(?:"([^"]*)",?\s*)')


def etag_cache(key=None):
    """Use the HTTP Entity Tag cache for Browser side caching

    If a "If-None-Match" header is found, and equivilant to ``key``,
    then a ``304`` HTTP message will be returned with the ETag to tell
    the browser that it should use its current cache of the page.

    Otherwise, the ETag header will be added to the response headers.
    """
    if_none_matches = IF_NONE_MATCH.findall(
        tg_request.environ.get("HTTP_IF_NONE_MATCH", "")
    )
    response = tg_response._current_obj()
    response.etag = key
    if str(key) in if_none_matches:
        response.headers.pop("Content-Type", None)
        response.headers.pop("Cache-Control", None)
        response.headers.pop("Pragma", None)
        raise status_map[304]()


def abort(
    status_code=None,
    detail="",
    headers=None,
    comment=None,
    passthrough=False,
    error_handler=False,
):
    """Aborts the request immediately by returning an HTTP exception

    In the event that the status_code is a 300 series error, the detail
    attribute will be used as the Location header should one not be
    specified in the headers attribute.

    **passthrough**
        When ``True`` instead of displaying the custom error
        document for errors or the authentication page for
        failed authorizations the response will just pass
        through as is.

        Set to ``"json"`` to send out the response body in
        JSON format.

    **error_handler**
        When ``True`` instead of immediately abort the request
        it will create a callable that can be used as ``@validate``
        error_handler.

        A common case is ``abort(404, error_handler=True)`` as
        ``error_handler`` for validation that retrieves objects
        from database::

            from formencode.validators import Wrapper

            @validate({'team': Wrapper(to_python=lambda value:
                                        Group.query.find({'group_name': value}).one())},
                      error_handler=abort(404, error_handler=True))
            def view_team(self, team):
                return dict(team=team)

    """
    exc = status_map[status_code](detail=detail, headers=headers, comment=comment)

    if passthrough == "json":
        exc.content_type = "application/json"
        exc.charset = "utf-8"
        exc.body = json_encode(dict(status=status_code, detail=str(exc))).encode(
            "utf-8"
        )

    def _abortion(*args, **kwargs):
        if passthrough:
            tg_request.environ["tg.status_code_redirect"] = False
            tg_request.environ["tg.skip_auth_challenge"] = True
        raise exc

    if error_handler is False:
        return _abortion()
    else:
        return _abortion


def validation_errors_response(*args, **kwargs):
    """Returns a :class:`.Response` object with validation errors.

    The response will be created with a *412 Precondition Failed*
    status code and errors are reported in JSON format as response body.

    Typical usage is as ``error_handler`` for JSON based api::

        @expose('json')
        @validate({'display_name': validators.NotEmpty(),
                   'group_name': validators.NotEmpty()},
                  error_handler=validation_errors_response)
        def post(self, **params):
            group = Group(**params)
            return dict(group=group)

    """
    req = tg_request._current_obj()
    errors = dict(
        (str(key), str(error)) for key, error in req.validation.errors.items()
    )
    values = req.validation.values
    try:
        return Response(status=412, json_body={"errors": errors, "values": values})
    except TypeError:
        # values cannot be encoded to JSON, this might happen after
        # validation passed and validators converted them to complex objects.
        # In this case use request params, instead of controller params.
        return Response(
            status=412, json_body={"errors": errors, "values": req.args_params}
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

from ...configuration.utils import TGConfigError
from ...support.converters import asbool, aslogger
from ...util.bunch import get_partial_dict
from ..base import (
    AppReadyConfigurationAction,
    BeforeConfigConfigurationAction,
    ConfigurationComponent,
)


class SimpleAuthenticationConfigurationComponent(ConfigurationComponent):
    """Provide support for Simple Authentication.

    Simple Authentication is the standard way to handle authentication
    and authorization in TurboGears. Where every request has an associated
    User object (which might be None) and each user can be part of one or more
    Groups each having a set of Permissions.

    The simple auth is based on ``repoze.who`` and by default sets the
    required ``authenticators``, ``identifiers`` and ``metadata providers``
    for a form based login. The user, its groups and permissions are retrieved
    through the ``authmetadata`` object configured by the application.

    For most cases instead of tweaking the simple authentication options
    you probably just want to change the behaviour of ``authmetadata``
    object in your application configuration.

    Provided options:

        * ``sa_auth.enabled``: Enabled/Disable simple authentication, by default True.
        * ``skip_authentication``: Disable authentication for tests, the
          user will always be authenticated through a ``REMOTE_USER`` environ
          key which will be considered the authenticated user id when set.
        * ``sa_auth.authmetadata``: authmetadata instance to use as an authenticator.
          Can be None for no authenticator.
        * ``sa_auth.log_stream``: Provide a custom logger fo authentication.
          by default the ``auth`` logger is used.
        * ``sa_auth.identifiers``: The identifiers to use to recognise a logged user.
          By default ``('default', None)`` leads to an authentication cookie being
          used to recognise logged users.
        * ``sa_auth.form_identifies``: Whenever to add the form plugin to
          the identifiers. By default ``True``. This allows the form plugin
          to intercept requests and identify the user credentisl from the
          data submitted by the login page. If this is
          disabled, the form will be able to act as a challenger and redirect
          the user to the login page, but it won't be able to actually get the
          user_name/password from the submitted form and provide them to the
          authenticator.
        * ``sa_auth.cookie_secret``: Secret to encode the ``auth_tkt`` cookie.
          This is only required when ``('default', None)`` is listed
          in ``sa_auth.identifiers``.
        * ``sa_auth.authenticators``: List of authenticators used to authenticate
          user against the provided credentials. By default ``('default', None)`` and
          ``cookie`` are the only enabled ones. This means that user will be
          authenticated against an user name and password using authmetadata or
          will be authenticated through the presence of an authentication cookie.
        * ``sa_auth.cookie_name``: Name of the cookie used to authenticate the
          user if the cookie identifer and authenticator are enabled (by default they are).
        * ``sa_auth.login_url``: Url where the login form is displayed if
          ``sa_auth.form_plugin`` is enabled. By default ``"/login"``.
        * ``sa_auth.login_handler``: Url that should handle form submitted authentication
          requests if ``sa_auth.form_plugin`` is enabled and it is allowed to identify.
          By default it's ``"/login_handler"``.
        * ``sa_auth.logout_handler``: Url that should handle logout requests
          if ``sa_auth.form_plugin`` is enabled and it is allowed to identify.
          By default it's ``"/logout_handler"``
        * ``sa_auth.post_login_url``: Where to redirect user after a login.
          Only applied if ``sa_auth.form_plugin`` is enabled and allowed to identify.
        * ``sa_auth.post_logout_url`` Where to redirect user after a logout.
          Only applied if ``sa_auth.form_plugin`` is enabled and allowed to identify.
        * ``sa_auth.login_counter_name``: Parameter used by form login to keep track
          of login attempts if `sa_auth.form_plugin`` is enabled. By default ``__logins``.
        * ``sa_auth.form_plugin``: Provide an alternative login/logout implement
          for form based authentication.  This might make all ``form_plugin`` related
          options unusable.
        * ``sa_auth.mdproviders``: Enable some metadata providers. This are used to
          inject additional user details into the current request. By default it's
          disabled and :class:`tg.appwrappers.identity.IdentityApplicationWrapper` is
          used instead.
        * ``sa_auth.authtkt.timeout``: Timeout of the authtkt token.
        * ``sa_auth.authtkt.reissue_time``: Frequency at which the authtkt token
                                            should be refreshed.
        * ``sa_auth.authtkt.secure``: Make cookie HTTPS only and disable Javascript access.
        * ``sa_auth.authtkt.samesite``: Enable a ``SameSite=X`` option on the cookie.

    """

    id = "sa_auth"

    def get_defaults(self):
        return {
            "sa_auth.enabled": False,
            "skip_authentication": False,
            "sa_auth.log_stream": logging.getLogger("auth"),
            "sa_auth.form_plugin": None,
        }

    def get_coercion(self):
        return {
            "sa_auth.enabled": asbool,
            "skip_authentication": asbool,
            "sa_auth.log_stream": aslogger,
            "sa_auth.form_identifies": asbool,
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator

        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError(
                "Simple Authentication only works on an ApplicationConfigurator"
            )

        from ...appwrappers.identity import IdentityApplicationWrapper

        configurator.register_application_wrapper(
            IdentityApplicationWrapper, after=True
        )

    def _configure(self, conf, app):
        if not conf["sa_auth.enabled"]:
            return

        if not conf["skip_authentication"] and "sa_auth.cookie_secret" not in conf:
            raise TGConfigError(
                "You must provide a value for authentication cookies secret. "
                "Make sure that you have an 'sa_auth.cookie_secret' config value."
            )

    def _add_middleware(self, conf, app):
        """
        Configure authentication and authorization.
        """
        # Start with the current configured authentication options.
        # Depending on the auth backend a new auth_args dictionary
        # can replace this one later on.
        if not conf["sa_auth.enabled"]:
            return app

        auth_args = get_partial_dict("sa_auth", conf)

        # Removing keywords not used by repoze.who:
        auth_args.pop("enabled", None)
        auth_args.pop("password_encryption_method", None)

        # Removing authmetadata as is not used by repoze.who:
        tgauthmetadata = auth_args.pop("authmetadata", None)

        try:
            pos = auth_args["authenticators"].index(("default", None))
        except KeyError:
            # Didn't specify authenticators, setup default one
            pos = None
        except ValueError:
            # Specified authenticators and default is not in there
            # so we want to skip default TG auth configuration.
            pos = -1

        if pos is None or pos >= 0:
            authenticator = None
            if tgauthmetadata:
                if not hasattr(tgauthmetadata, "authenticate"):
                    raise TGConfigError("authmetadata in app_cfg missing authenticate.")

                if tgauthmetadata.authenticate is not None:
                    from tg.configuration.auth import create_default_authenticator

                    auth_args, tgauth = create_default_authenticator(
                        tgauthmetadata, **auth_args
                    )
                    authenticator = ("tgappauth", tgauth)

                    if pos is None:
                        auth_args["authenticators"] = [authenticator]
                    else:
                        # We make a copy so that we don't modify the original one.
                        auth_args["authenticators"] = auth_args["authenticators"]
                        auth_args["authenticators"][pos] = authenticator

        from tg.configuration.auth import setup_auth

        app = setup_auth(
            app, skip_authentication=conf["skip_authentication"], **auth_args
        )
        return app



from __future__ import absolute_import

import os

from markupsafe import Markup

from tg.render import cached_template

from ..configuration.utils import coerce_config
from ..i18n import ugettext
from ..support.converters import asbool, aslist
from .base import RendererFactory

try:
    import kajiki
except ImportError:  # pragma: no cover
    kajiki = None

if kajiki is not None:
    from kajiki.loader import FileLoader
else:  # pragma: no cover

    class FileLoader(object):
        pass


__all__ = ["KajikiRenderer"]


class KajikiRenderer(RendererFactory):
    """
    Configuration Options available as ``templating.kajiki.*``:

        - ``templating.kajiki.force_mode`` -> Kajiki Rendering Mode (html, html5, xml). Default ``html5``.
        - ``templating.kajiki.template_extension`` -> Kajiki Templates extension, default ``.xhtml``
        - ``templating.kajiki.xml_autoblocks`` -> List of tags that should be automatically converted to blocks.
        - ``templating.kajiki.cdata_scripts`` -> Automatically wrap scripts in CDATA.
        - ``templating.kajiki.html_optional_tags`` -> Allow unclosed html, head and body tags.
        - ``templating.kajiki.strip_text`` -> Strip leading/trailing spaces from text nodes.

    Supported ``render_params``:

        - Caching options supported by :func:`.cached_template`
        - All arguments supported by :func:`kajiki.xml_template.XMLTemplate`

    """

    CONFIG_OPTIONS = {
        "force_mode": str,
        "template_extension": str,
        "autoescape_text": asbool,
        "xml_autoblocks": aslist,
        "cdata_scripts": asbool,
        "html_optional_tags": asbool,
        "strip_text": asbool,
    }
    engines = {"kajiki": {"content_type": "text/html"}}

    @classmethod
    def create(cls, config, app_globals):
        """Setup a renderer and loader for the Kajiki engine."""
        if kajiki is None:  # pragma: no cover
            return None

        options = coerce_config(config, "templating.kajiki.", cls.CONFIG_OPTIONS)
        if not options.get("html_optional_tags", False):
            # Kajiki by default doesn't close BODY and HEAD tags, but this behaviour
            # breaks TW2 resources that get injected at the end of head or body.
            # So by default we force Kajiki to close those tags.
            from kajiki import html_utils

            html_utils.HTML_OPTIONAL_END_TAGS.discard("html")
            html_utils.HTML_OPTIONAL_END_TAGS.discard("head")
            html_utils.HTML_OPTIONAL_END_TAGS.discard("body")

        # This is official way to switch gettext function in kajiki
        # as documented in http://pythonhosted.org/Kajiki/i18n.html
        from kajiki import i18n

        i18n.gettext = ugettext

        loader = KajikiTemplateLoader(
            config["paths"].templates[0],
            dotted_finder=app_globals.dotted_filename_finder,
            reload=config["auto_reload_templates"],
            **options,
        )
        return {"kajiki": cls(loader)}

    def __init__(self, loader):
        self.loader = loader

    def __call__(
        self,
        template_name,
        template_vars,
        cache_key=None,
        cache_type=None,
        cache_expire=None,
        **render_params,
    ):
        """Render a template with Kajiki

        Accepts the cache options ``cache_key``, ``cache_type``, and
        ``cache_expire``.

        """

        # Create a render callable for the cache function
        def render_template():
            # Grab a template reference
            template = self.loader.load(template_name, **render_params)
            return Markup(template(template_vars).render())

        return cached_template(
            template_name,
            render_template,
            cache_key=cache_key,
            cache_type=cache_type,
            cache_expire=cache_expire,
        )


class KajikiTemplateLoader(FileLoader):
    """Kaijik template loader supporting dotted filenames.
    Solves also the issue of not supporting relative paths when using
    py:extends in Kaijiki
    """

    def __init__(self, base, dotted_finder, reload=True, force_mode="html5", **kwargs):
        self.dotted_finder = dotted_finder
        self.template_extension = kwargs.pop("template_extension", ".xhtml")

        super(KajikiTemplateLoader, self).__init__(base, reload, force_mode, **kwargs)

    def _filename(self, filename):
        if not filename.endswith(self.template_extension):
            finder = self.dotted_finder
            filename = finder.get_dotted_filename(
                template_name=filename, template_extension=self.template_extension
            )

            if not os.path.exists(filename):
                raise IOError("Template %s not found" % filename)

        resolved_filename = super(KajikiTemplateLoader, self)._filename(filename)
        if resolved_filename is None:
            raise IOError("Template %s not found in template paths" % filename)
        return resolved_filename



# -*- coding: utf-8 -*-
from logging import getLogger

from ...support.converters import asbool
from ..base import (
    AppReadyConfigurationAction,
    BeforeConfigConfigurationAction,
    ConfigurationComponent,
)

log = getLogger(__name__)


class StaticsConfigurationComponent(ConfigurationComponent):
    """Provide support for serving Static Files.

    In production, use Apache or another web server to serve static files.

        - ``serve_static``: Enable / Disable serving static files. **Can be set from .ini file**
        - ``paths.static_files``: Directory where the static files should be served from.
                                  Refer to :class:`.PathsConfigurationComponent` for configuration.
    """

    id = "statics"

    def get_defaults(self):
        return {"serve_static": True}

    def get_coercion(self):
        return {"serve_static": asbool}

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def _configure(self, conf, app):
        if not conf["paths"]["static_files"]:
            conf["serve_static"] = False

    def _add_middleware(self, conf, app):
        if not conf["serve_static"]:
            return app

        from tg.support.statics import StaticsMiddleware

        return StaticsMiddleware(app, conf["paths"]["static_files"])



# -*- coding: utf-8 -*-
import logging

from .application import ApplicationConfigurator
from .components.app_globals import AppGlobalsConfigurationComponent
from .components.dispatch import DispatchConfigurationComponent
from .components.helpers import HelpersConfigurationComponent
from .components.mimetypes import MimeTypesConfigurationComponent
from .components.paths import PathsConfigurationComponent
from .components.registry import RegistryConfigurationComponent
from .components.rendering import TemplateRenderingConfigurationComponent
from .components.validation import ValidationConfigurationComponent

log = logging.getLogger(__name__)


class MinimalApplicationConfigurator(ApplicationConfigurator):
    """An ApplicationConfigurator that enables minimum set of components.

    This is meant to create small self contained applications that
    might serve the purpose of simple JSON webservices or micro
    web applications.

    Enables components for:

        - mimetypes
        - load controllers and templates from paths
        - dispatch requests to a root controller
        - provide tg.app_globals
        - provide helpers in templates
        - support templates rendering
        - enable requests local registry for tg.request, tg.response, etc...

    """

    def __init__(self):
        super(MinimalApplicationConfigurator, self).__init__()

        self.register(MimeTypesConfigurationComponent, after=False)
        self.register(PathsConfigurationComponent, after=False)
        self.register(DispatchConfigurationComponent, after=False)
        self.register(AppGlobalsConfigurationComponent)
        self.register(HelpersConfigurationComponent)
        self.register(TemplateRenderingConfigurationComponent)
        self.register(RegistryConfigurationComponent, after=True)
        self.register(ValidationConfigurationComponent)



"""
This is the main dispatcher module.

Dispatch works as follows:
Start at the RootController, the root controller must
have a _dispatch function, which defines how we move
from object to object in the system.
Continue following the dispatch mechanism for a given
controller until you reach another controller with a
_dispatch method defined.  Use the new _dispatch
method until anther controller with _dispatch defined
or until the url has been traversed to entirety.

This module also contains the standard ObjectDispatch
class which provides the ordinary TurboGears mechanism.

"""

import mimetypes as default_mimetypes
import weakref

from crank.dispatchstate import DispatchState
from webob.exc import HTTPException

import tg
from tg.caching import cached_property
from tg.request_local import WebObResponse

from ..wsgiapp import TGApp


def dispatched_controller():
    state = tg.request._dispatch_state
    for location, cont in reversed(state.controller_path):
        if cont.mount_point:
            return cont


class CoreDispatcher(object):
    """Extend this class to define your own mechanism for dispatch."""

    _use_lax_params = True
    _use_index_fallback = False

    def _get_dispatchable(self, context, url_path):
        """
        Returns a :class:`DispatchState` instance.

        :param context: The Request context.
        :param url_path: The url to dispatch.
        """
        req = context.request
        conf = context.config

        enable_request_extensions = not conf.get("disable_request_extensions", False)
        dispatch_path_translator = conf.get("dispatch_path_translator", True)

        state = DispatchState(
            weakref.proxy(req),
            self,
            req.args_params,
            url_path.split("/"),
            conf.get("ignore_parameters", []),
            strip_extension=enable_request_extensions,
            path_translator=dispatch_path_translator,
        )

        if enable_request_extensions:
            try:
                mimetypes = conf["mimetypes"]
            except KeyError:
                mimetypes = default_mimetypes

            ext = state.extension
            if ext is not None:
                ext = "." + ext
                mime_type, encoding = mimetypes.guess_type("file" + ext)
                req._fast_setattr("_response_type", mime_type)
            req._fast_setattr("_response_ext", ext)

        state = state.resolve()

        # Save the dispatch state for possible use within the controller methods
        req._fast_setattr("_dispatch_state", state)

        if conf.get("enable_routing_args", False):
            state.routing_args.update(state.params)
            if hasattr(state.root_dispatcher, "_setup_wsgiorg_routing_args"):
                state.root_dispatcher._setup_wsgiorg_routing_args(
                    state.path, state.remainder, state.routing_args
                )

        return state

    def _enter_controller(self, state, remainder):
        if hasattr(state.controller, "_visit"):
            state.controller._visit(*remainder, **state.params)

        return super(CoreDispatcher, self)._enter_controller(state, remainder)

    def _perform_call(self, context):
        """
        This function is called by __call__ to actually perform the controller
        execution.
        """
        py_request = context.request

        state = self._get_dispatchable(context, py_request.quoted_path_info)
        controller, action = state.controller, state.action
        params, remainder = state.params, state.remainder

        if hasattr(controller, "_before"):
            controller._before(*remainder, **params)

        self._setup_wsgi_script_name(state.path, remainder, params)

        r = self._call(action, params, remainder=remainder, context=context)

        if hasattr(controller, "_after"):
            controller._after(*remainder, **params)

        return r

    def __call__(self, environ, context):
        py_response = context.response

        try:
            response = self._perform_call(context)
        except HTTPException as httpe:
            response = httpe

        if response is tg.response or response is py_response:
            # Controller returned the response itself, so we need to do nothing.
            return response

        if response is None:
            # No content
            py_response.body = b""
            if py_response.status_int == 200:
                # Ensure that for missing content we return 'No Content', instead of 200 OK
                py_response.content_type = None
                py_response.status_int = 204
        elif isinstance(response, bytes):
            py_response.body = response
        elif isinstance(response, str):
            if not py_response.charset:
                py_response.charset = "utf-8"
            py_response.text = response
        elif isinstance(response, WebObResponse):
            # Copy eventual headers from tg.response
            for name, value in py_response.headers.items():
                header_name = name.lower()
                if header_name in ("content-type", "content-length"):
                    # Do not overwrite content related headers in returned response
                    continue
                if header_name == "set-cookie":
                    response.headers.add(name, value)
                else:
                    response.headers.setdefault(name, value)
            py_response = context.response = response
        else:
            py_response.app_iter = response

        return py_response

    @cached_property
    def mount_point(self):
        if not self.mount_steps:
            return ""
        return "/" + "/".join((x[0] for x in self.mount_steps[1:]))

    @cached_property
    def mount_steps(self):
        def find_url(root, item, parents):
            for i in dir(root):
                if i.startswith("_") or i in ("mount_steps", "mount_point"):
                    continue

                controller = getattr(root, i)
                if controller is item:
                    return parents + [(i, item)]
                if hasattr(controller, "_dispatch"):
                    v = find_url(
                        controller.__class__, item, parents + [(i, controller)]
                    )
                    if v:
                        return v
            return []

        root_controller = tg.config.get("tg.root_controller")
        if root_controller is None:
            root_controller = TGApp.lookup_controller(tg.config, "root")

        return find_url(root_controller, self, [("/", root_controller)])



# -*- coding: utf-8 -*-
import os

from tg.util import Bunch

from ..base import BeforeConfigConfigurationAction, ConfigurationComponent


class PathsConfigurationComponent(ConfigurationComponent):
    """Configure application paths.

    States where the application is contained, where
    to load controllers from, templates from, static files
    and which is the application Python package.

    Options:

        * ``package``: The python package containing the web application.
        * ``paths``: A Dictionary of directories where templates, static files
                     and controllers can be found::

                        {
                            'controllers': 'my/path/to/controlllers',
                            'static_files': 'my/path/to/files',
                            'templates': ['list/of/paths/to/templates']
                        )
        * ``static_files``: An alias to ``paths['static_files']`` for convenience.
    """

    id = "paths"

    DEFAULT_PATHS = {
        "root": None,
        "controllers": None,
        "templates": ["."],
        "static_files": None,
    }

    def get_defaults(self):
        return {"paths": Bunch(), "package": None}

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_paths),
            BeforeConfigConfigurationAction(self._configure_root_modules),
        )

    def _configure_paths(self, conf, app):
        try:
            package = conf["package"].__name__
        except (KeyError, AttributeError):
            package = None

        if package is not None:
            conf["package_name"] = conf["package"].__name__
        else:
            conf["package_name"] = None

        if package is None:
            # if we don't have a specified package (ES: we are in minimal mode)
            # we are unable to get paths from the package itself.
            paths = Bunch()
        else:
            root = os.path.dirname(os.path.abspath(conf["package"].__file__))
            paths = Bunch(
                root=root,
                controllers=os.path.join(root, "controllers"),
                static_files=os.path.join(root, "public"),
                templates=[os.path.join(root, "templates")],
            )

        # If the user defined custom paths, then use them instead of the
        # default ones:
        paths.update(conf["paths"])

        # Ensure all paths are set, load default ones otherwise
        for key, val in self.DEFAULT_PATHS.items():
            paths.setdefault(key, val)

        conf["paths"] = paths

    def _configure_root_modules(self, conf, app):
        root_module_path = conf["paths"]["root"]
        if not root_module_path:
            return None

        base_controller_path = conf["paths"]["controllers"]
        controller_path = base_controller_path[len(root_module_path) + 1 :]
        root_controller_module = ".".join(
            [conf["package_name"]] + controller_path.split(os.sep) + ["root"]
        )
        conf["application_root_module"] = root_controller_module



from __future__ import absolute_import

from ..request_local import Response, config, context, request
from ..support.registry import Registry
from ..wsgiapp import TGApp


def test_context(app, url=None, environ=None):
    """Sets up a TurboGears context allowing to use turbogears functions outside of a real request.

    Everything inside the with statement will have a TurboGears context available

    In case ``app`` is a ``TGApp`` instance (as the one that can be retrieved
    through `configure_new_app` hook) a context for such application is set in place.

    In case of a WebTest application a request to ``/_test_vars`` is performed
    to setup the test variables.

    In case ``app`` is ``None`` a fake application is created for which to setup the context.

    ``url`` parameter is provided to simulate the context as being for that url and
    ``environ`` parameter is provided to allow changing WSGI environ entries for the
    context.
    """
    if app is None:
        app = _BareTGAppMaker().make()

    url = url or "/"

    from webob.request import environ_from_url

    wsgienviron = environ_from_url(url)
    wsgienviron.update(environ or {})

    if isinstance(app, TGApp):
        return _TGTestContextManager(app, wsgienviron)
    else:
        return _WebTestTGTestContextManager(app, wsgienviron)


test_context.__test__ = False  # Prevent nose from detecting this as a test.


class _BareTGAppMaker(object):
    """Makes a new TGApp and returns it without any surrounding middleware."""

    def __init__(self):
        self.app = None

    def set_app(self, app):
        self.app = app
        return app

    def make(self):
        from ..configurator import MinimalApplicationConfigurator

        configurator = MinimalApplicationConfigurator()
        configurator.update_blueprint(
            {"root_controller": lambda *args: Response("HELLO")}
        )
        configurator.make_wsgi_app(wrap_app=self.set_app)
        return self.app


class _TGTestContextManager(object):
    """Provides a context manager to setup thread local context for a TGApp"""

    def __init__(self, app, environ):
        self._app = app
        self._environ = environ
        self._registry = Registry()
        self._registry.prepare()
        self._environ["paste.registry"] = self._registry

    def __enter__(self):
        self._app._setup_app_env(self._environ)
        return self._app

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._registry.cleanup()


class _WebTestTGTestContextManager(object):
    """Provides a context manager to setup thread local context for a WebTest TG app."""

    def __init__(self, app, environ):
        self._app = app
        self._environ = environ

    def __enter__(self):
        self._app.get("/_test_vars")
        request.environ.update(self._environ)
        return self._app

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            config._pop_object()
        except AssertionError:
            pass

        try:
            context._pop_object()
        except AssertionError:
            pass



# -*- coding: utf-8 -*-
import logging

from tg.configuration.utils import coerce_config
from tg.support.converters import asbool, asint, aslist

from ..base import (
    AppReadyConfigurationAction,
    BeforeConfigConfigurationAction,
    ConfigurationComponent,
)

log = logging.getLogger(__name__)


class SlowRequestsConfigurationComponent(ConfigurationComponent):
    """Provides slow requests reporting for TurboGears through BackLash.

    This is enabled through the ``trace_slowreqs.enable`` option and
    is only enabled when ``debug=false``.

    All the options available for error reporting are configured
    as ``trace_slowreqs.*`` options in your ``app_cfg`` or ``.ini`` files:

        - ``trace_slowreqs.enable`` -> Enable/Disable slow requests reporting,
          by default it's disabled.
        - ``trace_slowreqs.interval`` -> Report requests slower than this value (default: 25s)
        - ``trace_slowreqs.exclude`` -> List of urls that should be excluded

    Slow requests are reported using *EMail* or *Sentry*, the same
    options available in :class:`.ErrorReportingConfigurationComponent` apply
    with ``trace_slowreqs.`` instead of ``trace_errors.``.

    """

    id = "slow_requests"

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_backlash),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def _configure_backlash(self, conf, app):
        slowreqsware = coerce_config(
            conf,
            "trace_slowreqs.",
            {
                "smtp_use_tls": asbool,
                "dump_request_size": asint,
                "dump_request": asbool,
                "dump_local_frames": asbool,
                "dump_local_frames_count": asint,
                "enable": asbool,
                "interval": asint,
                "exclude": aslist,
            },
        )

        slowreqsware.setdefault("error_subject_prefix", "Slow Request: ")
        slowreqsware.setdefault("error_message", "A request is taking too much time")

        errorware = conf.get("tg.errorware", {})
        for erroropt in errorware:
            slowreqsware.setdefault(erroropt, errorware[erroropt])
        conf["tg.slowreqs"] = slowreqsware

    def _add_middleware(self, conf, app):
        errorware = conf["tg.slowreqs"]
        if errorware.get("enable") and not asbool(conf.get("debug")):
            reporters = []

            if errorware.get("error_email"):
                from backlash.tracing.reporters.mail import EmailReporter

                reporters.append(EmailReporter(**errorware))

            if errorware.get("sentry_dsn"):
                from backlash.tracing.reporters.sentry import SentryReporter

                reporters.append(SentryReporter(**errorware))

            if errorware.get("reporters", []):
                for reporter in errorware["reporters"]:
                    reporters.append(reporter)

            try:
                import backlash
            except ImportError:  # pragma: no cover
                log.warning(
                    "backlash not installed, slow requests reporting won't be available"
                )
            else:
                return backlash.TraceSlowRequestsMiddleware(
                    app,
                    reporters,
                    interval=errorware.get("interval", 25),
                    exclude_paths=errorware.get("exclude", None),
                    context_injectors=[_turbogears_backlash_context],
                )

        return app


def _turbogears_backlash_context(environ):
    tgl = environ.get("tg.locals")
    return {"request": getattr(tgl, "request", None)}



# -*- coding: utf-8 -*-
import os
from logging import getLogger

from ...configuration.utils import TGConfigError
from ...support.converters import aslist
from ..base import BeforeConfigConfigurationAction, ConfigurationComponent

log = getLogger(__name__)


class I18NConfigurationComponent(ConfigurationComponent):
    """Enable support for internationalization.

    Supported Options:

        * ``i18n.lang``: Default web app language if none was requested or
                         detected from browser.
                         This will always be listed as fallback language.
        * ``i18n.enabled``: Enable support for translations.
                            Disabling i18n will speedup requests and all
                            pages will be served as they were in ``i18n.lang`` value.
        * ``i18n.native``: Native languages templates are written into.
                           No translation will occurr for those language.
                           Provide space separated or as a list.
        * ``localedir``: Where to find translation catalogs.
                         By default it's project root/i18n

    Refer to :class:`.I18NApplicationWrapper` for additional options
    in supporting i18n.
    """

    id = "i18n"

    def get_defaults(self):
        return {"i18n.lang": None, "i18n.native": None, "i18n.enabled": True}

    def get_coercion(self):
        return {"i18n.native": aslist}

    def get_actions(self):
        return (BeforeConfigConfigurationAction(self._configure),)

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator

        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError("I18N only works on an ApplicationConfigurator")

        from ...appwrappers.i18n import I18NApplicationWrapper

        configurator.register_application_wrapper(I18NApplicationWrapper, after=True)

    def _configure(self, conf, app):
        if conf["paths"]["root"]:
            conf["localedir"] = os.path.join(conf["paths"]["root"], "i18n")
        else:
            conf["i18n.enabled"] = False



from __future__ import absolute_import

from os.path import exists, getmtime

from markupsafe import Markup

from tg.i18n import ugettext, ungettext
from tg.render import cached_template

from .base import RendererFactory

try:
    import jinja2
except ImportError:  # pragma: no cover
    jinja2 = None

if jinja2 is not None:
    from jinja2 import ChoiceLoader, Environment
    from jinja2.exceptions import TemplateNotFound
    from jinja2.filters import FILTERS
    from jinja2.loaders import FileSystemLoader
else:  # pragma: no cover

    class FileSystemLoader(object):
        pass


__all__ = ["JinjaRenderer"]


class JinjaRenderer(RendererFactory):
    """
    Currently Jinja2 support uses a bunch of options from ``tg.config``
    and doesn't provide its own namespace.
    """

    engines = {"jinja": {"content_type": "text/html"}}

    @classmethod
    def create(cls, config, app_globals):
        """Setup a renderer and loader for Jinja2 templates."""
        if jinja2 is None:  # pragma: no cover
            return None

        if config.get("use_dotted_templatenames", True):
            TemplateLoader = DottedTemplateLoader
            template_loader_args = {"dotted_finder": app_globals.dotted_filename_finder}
        else:
            TemplateLoader = FileSystemLoader
            template_loader_args = {}

        if "jinja_extensions" not in config:
            config["jinja_extensions"] = []

        # Add i18n extension by default
        if "jinja2.ext.i18n" not in config["jinja_extensions"]:
            config["jinja_extensions"].append("jinja2.ext.i18n")

        if "jinja_filters" not in config:
            config["jinja_filters"] = {}

        loader = ChoiceLoader(
            [
                TemplateLoader(path, **template_loader_args)
                for path in config["paths"]["templates"]
            ]
        )

        jinja2_env = Environment(
            loader=loader,
            autoescape=True,
            auto_reload=config["auto_reload_templates"],
            extensions=config["jinja_extensions"],
        )

        # Try to load custom filters module under app_package.lib.templatetools
        try:
            if not config["package_name"]:
                raise AttributeError()

            filter_package = config["package_name"] + ".lib.templatetools"
            autoload_lib = __import__(filter_package, {}, {}, ["jinja_filters"])
            try:
                autoload_filters = dict(
                    map(
                        lambda x: (x, autoload_lib.jinja_filters.__dict__[x]),
                        autoload_lib.jinja_filters.__all__,
                    )
                )
            except AttributeError:  # pragma: no cover
                autoload_filters = dict(
                    filter(
                        lambda x: callable(x[1]),
                        autoload_lib.jinja_filters.__dict__.iteritems(),
                    )
                )
        except (ImportError, AttributeError):
            autoload_filters = {}

        # Add jinja filters
        filters = dict(FILTERS, **autoload_filters)
        filters.update(config["jinja_filters"])
        jinja2_env.filters = filters

        # Jinja's unable to request c's attributes without strict_c
        config["tg.strict_tmpl_context"] = True

        # Add gettext functions to the jinja environment
        jinja2_env.install_gettext_callables(ugettext, ungettext)

        return {"jinja": cls(jinja2_env)}

    def __init__(self, jinja2_env):
        self.jinja2_env = jinja2_env

    def __call__(
        self,
        template_name,
        template_vars,
        cache_key=None,
        cache_type=None,
        cache_expire=None,
    ):
        """Render a template with Jinja2

        Accepts the cache options ``cache_key``, ``cache_type``, and
        ``cache_expire``.

        """

        # Create a render callable for the cache function
        def render_template():
            # Grab a template reference
            template = self.jinja2_env.get_template(template_name)
            return Markup(template.render(**template_vars))

        return cached_template(
            template_name,
            render_template,
            cache_key=cache_key,
            cache_type=cache_type,
            cache_expire=cache_expire,
        )


class DottedTemplateLoader(FileSystemLoader):
    """Jinja template loader supporting dotted filenames. Based on Genshi Loader"""

    def __init__(self, *args, **kwargs):
        self.template_extension = kwargs.pop("template_extension", ".jinja")
        self.dotted_finder = kwargs.pop("dotted_finder")

        super(DottedTemplateLoader, self).__init__(*args, **kwargs)

    def get_source(self, environment, template):
        # Check if dottedname
        if not template.endswith(self.template_extension):
            # Get the actual filename from dotted finder
            finder = self.dotted_finder
            template = finder.get_dotted_filename(
                template_name=template, template_extension=self.template_extension
            )
        else:
            return FileSystemLoader.get_source(self, environment, template)

        # Check if the template exists
        if not exists(template):
            raise TemplateNotFound(template)

        # Get modification time
        mtime = getmtime(template)

        # Read the source
        fd = open(template, "rb")
        try:
            source = fd.read().decode("utf-8")
        finally:
            fd.close()

        return source, template, lambda: mtime == getmtime(template)



import logging
import sys

from ..configuration.utils import coerce_config
from ..support.converters import asbool, asint
from .base import ApplicationWrapper

log = logging.getLogger(__name__)


class AbortTransaction(Exception):
    def __init__(self, response):
        self.response = response


class TransactionApplicationWrapper(ApplicationWrapper):
    """Wraps the whole application in zope.transaction transaction manager and
    rollbacks transaction in case of crashes.

    Supported options which can be provided by config are:

        - ``tm.enabled``: Whenever the transaction manager is enabled or not.
        - ``tm.attempts``: Number of times the transaction should be retried if
          it fails (no retry by default)
        - ``tm.commit_veto``: A function that will be called for every transaction to check
          if it should abort transaction or let it go. Function signature should be:
          ``function(environ, status_code, headers) -> bool``.

    """

    def __init__(self, handler, config):
        super(TransactionApplicationWrapper, self).__init__(handler, config)

        options = {"enabled": False, "attempts": 1, "commit_veto": None}
        options.update(
            coerce_config(
                config,
                "tm.",
                {
                    "enabled": asbool,
                    "attempts": asint,
                },
            )
        )

        self.enabled = options["enabled"]
        self.attempts = options["attempts"]
        self.commit_veto = options["commit_veto"]
        self.manager = None

        if self.enabled:
            try:
                import transaction

                self.manager = transaction.manager
            except ImportError:  # pragma: no cover
                log.exception("Unable to Enable Transaction Manager")
                self.enabled = False

        log.debug("TransactionManager enabled: %s -> %s", self.enabled, options)

    @property
    def injected(self):
        return self.enabled

    def __call__(self, controller, environ, context):
        if "repoze.tm.active" in environ:  # pragma: no cover
            # Skip transaction manager if repoze.tm2 is enabled
            return self.next_handler(controller, environ, context)

        transaction_manager = self.manager
        total_attempts = self.attempts
        commit_veto = self.commit_veto

        attempts_left = total_attempts
        while attempts_left:
            attempts_left -= 1

            log.debug("Attempts Left %d (%d total)", attempts_left, total_attempts)
            transaction_manager.begin()
            try:
                response = self.next_handler(controller, environ, context)
                if transaction_manager.isDoomed():
                    log.debug("Transaction doomed")
                    raise AbortTransaction(response)

                if commit_veto is not None:
                    veto = commit_veto(environ, response.status, response.headerlist)
                    if veto:
                        log.debug("Transaction vetoed")
                        raise AbortTransaction(response)

                transaction_manager.commit()
                log.debug("Transaction committed!")
                return response
            except AbortTransaction as e:
                transaction_manager.abort()
                return e.response
            except Exception:
                exc_info = sys.exc_info()
                log.debug("Error while running request, aborting transaction")
                txn = transaction_manager.get()
                try:
                    try:
                        can_retry = txn.isRetryableError(exc_info[1])
                    except AttributeError:
                        can_retry = transaction_manager._retryable(*exc_info[:-1])
                    txn.abort()
                    if (attempts_left <= 0) or (not can_retry):
                        raise
                finally:
                    del exc_info



from ..request_local import request as tg_request
from ..util.urls import build_url


def url(base_url="/", params=None, qualified=False, scheme=None):
    """Generate an absolute URL that's specific to this application.

    The URL function takes a string (base_url) and, appends the
    SCRIPT_NAME and adds parameters for all of the
    parameters passed into the params dict.

    ``scheme`` can be passed in case of a ``qualified`` url to
    create an url with the given scheme.

    """
    if not isinstance(base_url, str) and hasattr(base_url, "__iter__"):
        base_url = "/".join(base_url)

    req = tg_request._current_obj()
    base_url = build_url(req.environ, base_url, params)
    if qualified:
        base_url = req.host_url + base_url
        if scheme is not None:
            base_url = scheme + base_url[len(req.scheme) :]

    return base_url


class LazyUrl(object):
    """
    Wraps tg.url in an object that enforces evaluation of the url
    only when you try to display it as a string.
    """

    def __init__(self, base_url, params=None, **kwargs):
        self.base_url = base_url
        self.params = params
        self.kwargs = kwargs
        self._decoded = None

    @property
    def _id(self):
        if self._decoded is None:
            self._decoded = url(self.base_url, params=self.params, **self.kwargs)
        return self._decoded

    @property
    def id(self):
        return self._id

    def __repr__(self):
        return self._id

    def __html__(self):
        return str(self)

    def __str__(self):
        return str(self._id)

    def encode(self, *args, **kw):
        return self._id.encode(*args, **kw)

    def __add__(self, other):
        return self._id + other

    def __radd__(self, other):
        return other + self._id

    def startswith(self, *args, **kw):
        return self._id.startswith(*args, **kw)

    def format(self, *args, **kwargs):
        return self._id.format(*args, **kwargs)

    def __json__(self):
        return str(self)


def lurl(base_url=None, params=None, **kwargs):
    """
    Like tg.url but is lazily evaluated.

    This is useful when creating global variables as no
    request is in place.

    As without a request it wouldn't be possible
    to correctly calculate the url using the SCRIPT_NAME
    this demands the url resolution to when it is
    displayed for the first time.
    """
    return LazyUrl(base_url, params, **kwargs)



# -*- coding: utf-8 -*-

from .application import ApplicationConfigurator
from .base import (
    AppReadyConfigurationAction,
    BeforeConfigConfigurationAction,
    ConfigReadyConfigurationAction,
    ConfigurationComponent,
    Configurator,
    EnvironmentLoadedConfigurationAction,
)
from .fullstack import FullStackApplicationConfigurator
from .minimal import MinimalApplicationConfigurator

__all__ = (
    "AppReadyConfigurationAction",
    "BeforeConfigConfigurationAction",
    "ConfigurationComponent",
    "ConfigReadyConfigurationAction",
    "Configurator",
    "EnvironmentLoadedConfigurationAction",
    "ApplicationConfigurator",
    "MinimalApplicationConfigurator",
    "FullStackApplicationConfigurator",
)



# -*- coding: utf-8 -*-
import os
from logging import getLogger

from ...configuration.utils import TGConfigError
from ..base import BeforeConfigConfigurationAction, ConfigurationComponent

log = getLogger(__name__)


class SessionConfigurationComponent(ConfigurationComponent):
    """Provide support for sessions through Beaker.

    Session components adds support for setting up the session
    manager used by ``tg.sessions``.

    Options:

        * ``session.enabled``: Add support for sessions in the application.
                               By default sesions are enabled.
        * ``session.data_dir``: Where to store session files, by default
                                the ``$cache_dir/sessions`` is used.

    Refer to :class:`.SessionApplicationWrapper` for all the supported
    options.

    """

    id = "session"

    def get_defaults(self):
        return {"session.enabled": True}

    def get_actions(self):
        return (BeforeConfigConfigurationAction(self._configure),)

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator

        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError("Sessions only work on an ApplicationConfigurator")

        from ...appwrappers.session import SessionApplicationWrapper

        configurator.register_application_wrapper(SessionApplicationWrapper, after=True)

    def _configure(self, conf, app):
        if "cache_dir" in conf:
            conf.setdefault(
                "session.data_dir", os.path.join(conf["cache_dir"], "sessions")
            )



import urllib.parse


def _smart_str(s):
    """
    Returns a bytestring version of 's', encoded as specified in 'encoding'.

    This function was borrowed from Django.

    """
    if isinstance(s, bytes):
        return s
    elif isinstance(s, str):
        return s.encode("utf-8", "strict")
    else:
        try:
            return str(s).encode("ascii", "strict")
        except UnicodeEncodeError:
            if isinstance(s, Exception):
                # An Exception subclass containing non-ASCII data that doesn't
                # know how to print itself properly. We shouldn't raise a
                # further exception.
                return " ".join(
                    [_smart_str(arg).decode("utf-8") for arg in s.args]
                ).encode("utf-8", "strict")
            return str(s).encode("utf-8", "strict")


def _generate_smart_str(params):
    for key, value in params.items():
        if value is None:
            continue
        if isinstance(value, (list, tuple)):
            for item in value:
                yield _smart_str(key), _smart_str(item)
        else:
            yield _smart_str(key), _smart_str(value)


def _urlencode(params):
    """
    A version of Python's urllib.urlencode() function that can operate on
    unicode strings. The parameters are first case to UTF-8 encoded strings and
    then encoded as per normal.
    """
    return urllib.parse.urlencode([i for i in _generate_smart_str(params)])


def build_url(environ, base_url="/", params=None):
    """Build a URL based on the given WSGI environ, a base URL and its parameters."""
    if base_url.startswith("/"):
        base_url = environ["SCRIPT_NAME"] + base_url

    if params:
        return "?".join((base_url, _urlencode(params)))

    return base_url



# -*- coding: utf-8 -*-
import logging
import weakref

from ..configuration.utils import (
    DependenciesList,
    DictionaryView,
    coerce_options,
    copyoption,
)

log = logging.getLogger(__name__)


class Configurator(object):
    """Manages a configuration process with multiple components and steps.

    Multiple registered components will be configured by applying
    the options provided during configuration on top of a bluepint
    (a set of default options).

    The result will be a configuration dictionary and whatever side effect
    each component triggered when being configured.

    This doesn't really know anything about that it is configuring
    and is not TurboGears specific. An ``ApplicationConfigurator``
    is provided which is a specialised ``Configurator`` for configuring
    TurboGears applications.

    .. versionadded:: 2.4
    """

    def __init__(self):
        self._initialized = False
        self._blueprint = {}
        self._coercion = {}
        self._components = DependenciesList()

    def _initialize(self):
        if self._initialized:
            return

        for _, component in self._components:
            component.on_bind(self)
        self._initialized = True

    def update_blueprint(self, config):
        """Add options from ``config`` to the configuration blueprint."""
        self._blueprint.update(config)

    def get_blueprint_value(self, name):
        """Get an option registered into the configuration blueprint."""
        try:
            return self._blueprint[name]
        except KeyError:
            raise KeyError(
                "Configuration Blueprint does not provide a '{0}' option".format(name)
            )

    def get_blueprint_view(self, key):
        """A view is a subset of the blueprint options.

        If the blueprint contains multiple values that share
        a common prefix it is possible to get a view that
        contains them all. A view acts like a dictionary
        where it is possible to get or set the values.

        Given a blueprint that contains ``section.option1`` and
        ``section.option2`` it is possible to ask for the view
        of ``section`` and get back a dictionary like object
        that contains ``option1`` and ``option2``.
        """
        if key.endswith("."):
            raise ValueError("A Blueprint key cannot end with a .")
        return DictionaryView(self._blueprint, key)

    def register(self, component_type, after=None):
        """Registers a new configuration component to be performed by the Configurator"""
        if not issubclass(component_type, ConfigurationComponent):
            raise ValueError(
                "Configuration component must inherit ConfigurationComponent"
            )

        component = component_type()
        component._prepare_blueprint(self._blueprint)
        component._prepare_coercion(self._coercion)
        self._components.add(component, key=component_type.id, after=after)

    def replace(self, component_id, new_component_type):
        """Replaces an existing configuration component with another one.

        Any default value set in the blueprint by the previous component
        won't be discarded and won't be replaced. So if you tuned the configuration
        of the previous component, the new one will preserve it.

        Also the id at which the component is registered won't be the one
        of the new component, but will be the one of the old component.

        In theory you should only replace components with components that
        provide same id and same default options.

        Note that replacing a component will only influence the applications
        created after it was replaced.
        """
        if not issubclass(new_component_type, ConfigurationComponent):
            raise ValueError(
                "Configuration component must inherit ConfigurationComponent"
            )

        component = new_component_type()
        component._prepare_blueprint(self._blueprint)
        component._prepare_coercion(self._coercion)
        self._components.replace(component_id, component)

    def get_component(self, component_id):
        """Retrieve a registered configuration component."""
        return self._components.get(component_id)

    def configure(self, global_conf=None, app_conf=None):
        """Prepare a configuration using the configurator.

        Once it's ready the configuration is returned.
        """
        self._initialize()

        conf = {}

        conf.update(copyoption(self._blueprint))
        conf.update(copyoption(global_conf))
        conf.update(app_conf)

        # Let track of the configurator that generated the configuration
        # into the configuration itself.
        conf["tg.configurator"] = weakref.ref(self)

        # Convert the loaded options according to the coercion functions
        # registered by each configuration component.
        conf.update(coerce_options(conf, self._coercion))

        for _, component in self._components:
            component._apply(BeforeConfigConfigurationAction, conf)

        log.debug("Initializing configuration, package: '%s'", conf.get("package_name"))
        return conf

    def setup(self, conf):
        """Given a configuration setup the environment.

        Applies ``ConfigReadyConfigurationAction`` configuration actions
        from the registered components.

        This usually involves all the configuration steps that already
        need to have all configuration options available before they can proceed.
        """
        for _, component in self._components:
            component._apply(ConfigReadyConfigurationAction, conf)


class ConfigurationComponent(object):
    """Represents a configuration component that will collaborate in configuring an application.

    Each component once registered in a :class:`.Configurator` can perform multiple
    actions at different times, each action can be registered by
    returing them through :meth:`.get_actions`.

    .. versionadded:: 2.4
    """

    def __init__(self):
        if not hasattr(self, "id"):
            raise ValueError(
                "ConfigurationComponent must provide an id class attribute "
                "to uniquely identify the component."
            )

        self._actions = {}
        for action in self.get_actions():
            self._register_action(action)

    def on_bind(self, configurator):
        """Called when component is bound to a Configurator.

        This should be overridden to register global settings and
        entities like ApplicationWrappers and RenderingEngines into
        the configurator. It won't be called more than once per
        Configurator.
        """
        return

    def get_actions(self):
        """Can be overridden to provide a set of actions the component has the perform.

        Must return an iterable containing all the actions that must be
        performed by this configuration component at any given time::

            (BeforeConfigConfigurationAction(self.setup_config),
             ConfigReadyConfigurationAction(self.initialize_things),
             EnvironmentLoadedConfigurationAction(self.add_middleware))
        """
        return tuple()

    def get_defaults(self):
        """Can be overridden to provide default values for configuration blueprint.

        This is used when the configuration component is added to a Configurator to
        retrieve any default value for configuration options.

        Must return a dictionary in the form::

            {'option_name': 'value'}

        """
        return {}

    def get_coercion(self):
        """Can be overridden to provide coercion methods for options.

        Must return a dictionary in the form::

            {'option_name': coerce_function}
        """
        return {}

    def _register_action(self, action):
        self._actions.setdefault(action.__class__.__name__, []).append(action)

    def _prepare_blueprint(self, blueprint):
        defaults = self.get_defaults()
        for k, v in defaults.items():
            blueprint.setdefault(k, v)

    def _prepare_coercion(self, coercion):
        defaults = self.get_coercion()
        for k, v in defaults.items():
            coercion.setdefault(k, v)

    def _apply(self, action_type, conf, app=None):
        for action in self._actions.get(action_type.__name__, []):
            log.debug("%s applying %s", self.__class__.__name__, action)
            app = action(conf, app)
        return app


class _ConfigurationAction(object):
    """An action done by a :class:`,ConfigurationComponent` during configuration process.

    Represents what should be done during part of configuration
    process by a :class:`ConfigurationComponent`.
    The :class:`.Configurator` will fire the actions at specific
    times based on the action type.

    .. versionadded:: 2.4
    """

    def __init__(self, perform=None):
        self.perform = perform

    def __repr__(self):
        return "<%s: %r>" % (self.__class__.__name__, self.perform)

    def __call__(self, conf, app):
        return self.perform(conf, app)


class BeforeConfigConfigurationAction(_ConfigurationAction):
    """An action to be executed before the app configuration is initialised."""

    pass


class ConfigReadyConfigurationAction(_ConfigurationAction):
    """An action to be executed once the configuration is loaded and ready."""

    pass


class EnvironmentLoadedConfigurationAction(_ConfigurationAction):
    """An action to be executed once the environment needed to create the app is ready."""

    pass


class AppReadyConfigurationAction(_ConfigurationAction):
    """An action to be executed once the application has been created.

    It's typically the best time where to wrap WSGI middlewares
    around the application.
    """

    pass



import logging
import random

import tg

try:
    from sqlalchemy.orm import Session
except ImportError:  # pragma: no cover

    class Session(object):
        """SQLAlchemy Session"""


log = logging.getLogger(__name__)


class BalancedSession(Session):
    _force_engine = None

    def get_bind(self, mapper=None, clause=None):
        config = tg.config._current_obj()

        engines = config.get("balanced_engines")
        if not engines:
            log.debug("Balancing disabled, using master")
            return config["tg.app_globals"].sa_engine

        forced_engine = self._force_engine
        if not forced_engine:
            try:
                forced_engine = tg.request._tg_force_sqla_engine
            except TypeError:
                forced_engine = "master"
            except AttributeError:
                pass

        if forced_engine:
            log.debug("Forced engine: %s", forced_engine)
            return engines["all"][forced_engine]
        elif self._flushing:
            log.debug("Choose engine: master")
            return engines["master"]
        else:
            choosen_slave = random.choice(list(engines["slaves"].keys()))
            log.debug("Choose engine: %s", choosen_slave)
            return engines["slaves"][choosen_slave]

    def using_engine(self, engine_name):
        return UsingEngineContext(engine_name, self)


class UsingEngineContext(object):
    def __init__(self, engine_name, DBSession=None):
        self.engine_name = engine_name
        if not DBSession:
            DBSession = tg.config["DBSession"]()
        self.session = DBSession
        self.past_engine = self.session._force_engine

    def __enter__(self):
        self.session._force_engine = self.engine_name
        return self.session

    def __exit__(self, exc_type, exc_value, traceback):
        self.session._force_engine = self.past_engine


def force_request_engine(engine_name):
    tg.request._tg_force_sqla_engine = engine_name



# -*- coding: utf-8 -*-

import logging

from webob.acceptparse import create_accept_header

from ..configuration import config, milestones

log = logging.getLogger(__name__)


def _decorated_controller_caller(tg_config, controller, remainder, params):
    try:
        application_controller_caller = tg_config["controller_caller"]
    except KeyError:  # pragma: no cover
        # This should never happen as controller_caller is setup by MinimalApplicationConfigurator.
        from tg.configurator.components.dispatch import _call_controller

        application_controller_caller = _call_controller

    return application_controller_caller(tg_config, controller, remainder, params)


class Decoration(object):
    """Simple class to support 'simple registration' type decorators"""

    def __init__(self, controller):
        self.controller = controller
        self.controller_caller = _decorated_controller_caller
        self._expositions = []
        self.engines = {}
        self.engines_keys = []
        self.default_engine = None
        self.custom_engines = {}
        self.render_custom_format = None
        self.validations = []
        self.inherit = False
        self.requirements = []
        self.hooks = dict(
            before_validate=[], before_call=[], before_render=[], after_render=[]
        )

    def __repr__(self):  # pragma: no cover
        return "<Decoration %s for %r>" % (id(self), self.controller)

    @classmethod
    def get_decoration(cls, func):
        try:
            dec = func.decoration
        except Exception:
            dec = func.decoration = cls(func)
        return dec

    def _register_exposition(self, exposition, inherit=False, before=False):
        """Register an exposition for later application"""

        # We need to store a reference to the exposition
        # so that we can merge them when inheritance is performed
        if before:
            self._expositions.insert(0, exposition)
        else:
            self._expositions.append(exposition)

        if inherit:
            # if at least one exposition is in inherit mode
            # all of them must inherit
            self.inherit = True

        milestones.renderers_ready.register(self._resolve_expositions)

    def _resolve_expositions(self):
        """Applies all the registered expositions"""
        while True:
            try:
                exposition = self._expositions.pop(0)
                exposition._apply()
            except IndexError:
                break

    @property
    def exposed(self):
        return bool(self.engines) or bool(self.custom_engines)

    def merge(self, deco):
        # This merges already registered template engines
        self.engines = dict(tuple(deco.engines.items()) + tuple(self.engines.items()))
        self.engines_keys = sorted(self.engines, reverse=True)
        self.custom_engines = dict(
            tuple(deco.custom_engines.items()) + tuple(self.custom_engines.items())
        )

        # This merges yet to register template engines
        for exposition in reversed(deco._expositions):
            self._register_exposition(exposition._clone(self.controller), before=True)

        # inherit all the parent hooks
        # parent hooks before current hooks so that they get called before
        for hook_name, hooks in deco.hooks.items():
            self.hooks[hook_name] = hooks + self.hooks[hook_name]

        # Inherit al validators registered on parent.
        self.validations = deco.validations + self.validations

    def register_template_engine(
        self, content_type, engine, template, exclude_names, render_params
    ):
        """Registers an engine on the controller.

        Multiple engines can be registered, but only one engine per
        content_type.  If no content type is specified the engine is
        registered at */* which is the default, and will be used
        whenever no content type is specified.

        exclude_names keeps track of a list of keys which will be
        removed from the controller's dictionary before it is loaded
        into the template.  This allows you to exclude some information
        from JSONification, and other 'automatic' engines which don't
        require a template.

        render_params registers extra parameters which will be sent
        to the rendering method.  This allows you to influence things
        like the rendering method or the injected doctype.

        """
        default_renderer = config.get("default_renderer")
        available_renderers = config.get("renderers", [])

        if engine and not available_renderers:
            log.warning(
                "Renderers not registered yet while exposing template %s for engine %s, "
                "skipping engine availability check",
                template,
                engine,
            )

        if engine and available_renderers and engine not in available_renderers:
            log.debug(
                "Registering template %s for engine %s not available. Skipping it",
                template,
                engine,
            )
            return

        content_type = content_type or "*/*"

        try:
            current_content_type_engine = self.engines[content_type][0]
        except (KeyError, IndexError):
            current_content_type_engine = None

        if current_content_type_engine is not None and engine != default_renderer:
            # Avoid overwriting the default renderer when there is already a template registered
            return

        self.engines[content_type] = (
            engine,
            template,
            exclude_names or [],
            render_params or {},
        )

        # Avoid engine lookup if we have only one engine registered
        if len(self.engines) == 1:
            self.default_engine = content_type
        else:
            self.default_engine = None

        # This is a hack to make text/html prominent in respect to other common choices
        # when they have the same weight for webob.acceptparse.Accept.best_match().
        # It uses the fact that the most common content types are all alphabetically
        # precedent to text/html, and so sorting engine keys alphabetically reversed
        # should make text/html the first choice when no other better choices are available.
        self.engines_keys = sorted(self.engines, reverse=True)

    def register_custom_template_engine(
        self,
        custom_format,
        content_type,
        engine,
        template,
        exclude_names,
        render_params,
    ):
        """Registers a custom engine on the controller.

        Multiple engines can be registered, but only one engine per
        custom_format.

        The engine is registered when @expose is used with the
        custom_format parameter and controllers render using this
        engine when the :meth:`use_custom_format` function is called
        with the corresponding custom_format.

        exclude_names keeps track of a list of keys which will be
        removed from the controller's dictionary before it is loaded
        into the template.  This allows you to exclude some information
        from JSONification, and other 'automatic' engines which don't
        require a template.

        render_params registers extra parameters which will be sent
        to the rendering method.  This allows you to influence things
        like the rendering method or the injected doctype.

        """

        self.custom_engines[custom_format or '"*/*"'] = (
            content_type,
            engine,
            template,
            exclude_names,
            render_params or {},
        )

    def lookup_template_engine(self, tgl):
        """Return the template engine data.

        Provides a convenience method to get the proper engine,
        content_type, template, and exclude_names for a particular
        tg_format (which is pulled off of the request headers).

        """
        request = tgl.request
        response = tgl.response

        try:
            render_custom_format = request._render_custom_format[self.controller]
        except Exception:
            render_custom_format = self.render_custom_format

        if render_custom_format:
            (content_type, engine, template, exclude_names, render_params) = (
                self.custom_engines[render_custom_format]
            )
        else:
            if self.default_engine:
                content_type = self.default_engine
            elif self.engines:
                if response.content_type is not None:
                    # Check for overridden content type from the controller call
                    accept_types = create_accept_header(response.content_type)
                elif request._response_type and request._response_type in self.engines:
                    # Check for content type detected by request extensions
                    accept_types = create_accept_header(request._response_type)
                else:
                    accept_types = request.accept

                best_matches = (
                    accept_types.acceptable_offers(self.engines_keys)
                    or
                    # If none of the available engines matches with the
                    # available options, just suggest usage of the first engine.
                    ((self.engines_keys[0], None),)
                )
                content_type = best_matches[0][0]
            else:
                content_type = "text/html"

            # check for overridden templates
            try:
                cnt_override_mapping = request._override_mapping[self.controller]
                engine, template, exclude_names, render_params = cnt_override_mapping[
                    content_type.split(";")[0]
                ]
            except (AttributeError, KeyError):
                (engine, template, exclude_names, render_params) = self.engines.get(
                    content_type, (None,) * 4
                )

        return content_type, engine, template, exclude_names, render_params

    def _register_hook(self, hook_name, func):
        """Registers the specified function as a hook.

        This is internal API which is used by tg.hooks, instead of
        calling this tg.hooks.register should be used.

        We now have four core hooks that can be applied by adding
        decorators: before_validate, before_call, before_render, and
        after_render. register_hook attaches the function to the hook
        which get's called at the appropriate time in the request life
        cycle.)
        """
        self.hooks.setdefault(hook_name, []).append(func)

    def _register_requirement(self, requirement):
        self._register_hook("before_call", requirement._check_authorization)
        self.requirements.append(requirement)

    def _register_controller_wrapper(self, wrapper):
        self.controller_caller = wrapper(self.controller_caller)

    def _register_validation(self, validation):
        self.validations.insert(0, validation)



# -*- coding: utf-8 -*-
from importlib import import_module
from logging import getLogger

from ...util import Bunch, DottedFileNameFinder
from ..base import ConfigReadyConfigurationAction, ConfigurationComponent

log = getLogger(__name__)


class AppGlobalsConfigurationComponent(ConfigurationComponent):
    """Enables the application global object.

    ``tg.app_globals`` is a global object always available in TG
    and shared across all requests and threads of the application.

    It's meant to keep around any value or method that all other
    parts of the application might depend on. As it's shared
    between all application threads and requests its content
    should be immutable or it will lead to race conditions.

    By default an instance of ``.lib.app_globals.Globals`` from
    within your application package is used. If it's not available
    an empty ``Bunch`` will be created instead.

    Provided options:

    * ``app_globals`` -> Factory or Class that should be used to create
      application global object instead of getting it from the application package.

    """

    id = "app_globals"

    def get_defaults(self):
        return {
            "app_globals": None,
        }

    def get_actions(self):
        return (ConfigReadyConfigurationAction(self._setup_app_globals),)

    def _setup_app_globals(self, conf, app):
        # Setup AppGlobals
        gclass = conf.pop("app_globals", None)
        if gclass is None:
            try:
                gclass = conf["package"].lib.app_globals.Globals
            except AttributeError:
                pass

        if gclass is None:
            try:
                app_globals_mod = import_module(
                    ".lib.app_globals", package=conf["package"].__name__
                )
                gclass = getattr(app_globals_mod, "Globals")
            except (ImportError, AttributeError):
                pass

        if gclass is None:
            log.warning(
                "app_globals not provided and lib.app_globals.Globals is not available."
            )
            gclass = Bunch

        g = gclass()

        g.dotted_filename_finder = DottedFileNameFinder()
        conf["tg.app_globals"] = g



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from ...configuration.utils import TGConfigError
from ...support.middlewares import DBSessionRemoverMiddleware
from ..base import (
    AppReadyConfigurationAction,
    ConfigReadyConfigurationAction,
    ConfigurationComponent,
)


class SQLAlchemyConfigurationComponent(ConfigurationComponent):
    """Support Relational Databases through SQLAlchemy.

    Configures SQLAlchemy connection to the database,
    the automatic session cleanup at the end of each request
    and support for master/slave databases.

    Support for transaction commit/rollback on request state
    is provided by :class:`.TransactionManagerConfigurationComponent`.

    The configured SQLAlchemy engine is made available as
    ``tg.app_globals.sa_engine``. While the configured SQLAlchemy
    session is made available as ``tg.config['SQLASession']``.

    Options:

        * ``use_sqlalchemy``: Enable SQLAlchemy in your application.
        * ``model``: Configure where models could be found, by default
                    they are expected in ``model`` Python module
                    within your application package.
        * ``sqlalchemy.*``: Options provided to SQLAlchemy connection
                           as expected by :func:`.engine_from_config`.
        * ``sqlalchemy.master.*`` Enable Master/Slave replication support
                                 and configure the master connection.
        * ``sqlalchemy.slaves.slavename.*``: Configure ``slavename`` slave in
                                            Master/Slave support and setup its
                                            connection. Multiple entries with
                                            different ``slavename`` values can
                                            be provided.

    """

    id = "sqlalchemy"

    def get_defaults(self):
        return {
            "use_sqlalchemy": False,
        }

    def get_actions(self):
        return (
            ConfigReadyConfigurationAction(self._setup_sqlalchemy),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator

        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError(
                "SQLAlchemy Support only works on an ApplicationConfigurator"
            )

    def _setup_sqlalchemy(self, conf, app):
        if not conf["use_sqlalchemy"]:
            return
        self.setup_sqlalchemy(conf, app)

    def setup_sqlalchemy(self, conf, app):
        """Setup SQLAlchemy database engine"""
        engine = self.create_sqlalchemy_engine(conf)

        conf["tg.app_globals"].sa_engine = engine

        model = self._get_model(conf)
        sqla_session = model.init_model(engine)
        if sqla_session is not None:
            # If init_model returns a specific session, keep it around
            # as the SQLAlchemy Session.
            conf["SQLASession"] = sqla_session

        if "DBSession" not in conf:
            # If the user hasn't specified a default session, assume
            # he/she uses the default DBSession in model
            conf["DBSession"] = model.DBSession

    def _add_middleware(self, conf, app):
        if not conf["use_sqlalchemy"]:
            return app
        return self.add_middleware(conf, app)

    def add_middleware(self, conf, app):
        """Set up middleware that cleans up the sqlalchemy session.

        The default behavior of TG 2 is to clean up the session on every
        request.  Only override this method if you know what you are doing!

        """
        dbsession = conf.get("SQLASession")
        if dbsession is None:
            dbsession = conf["DBSession"]
        return DBSessionRemoverMiddleware(dbsession, app)

    def _get_model(self, conf):
        try:
            package_models = conf["package"].model
        except AttributeError:
            package_models = None

        model = conf.get("model", package_models)
        if model is None:
            raise TGConfigError("SQLAlchemy enabled, but no models provided")

        return model

    @staticmethod
    def create_sqlalchemy_engine(conf):
        from sqlalchemy import engine_from_config

        balanced_master = conf.get("sqlalchemy.master.url")
        if not balanced_master:
            engine = engine_from_config(conf, "sqlalchemy.")
        else:
            engine = engine_from_config(conf, "sqlalchemy.master.")
            conf["balanced_engines"] = {
                "master": engine,
                "slaves": {},
                "all": {"master": engine},
            }

            all_engines = conf["balanced_engines"]["all"]
            slaves = conf["balanced_engines"]["slaves"]
            for entry in conf.keys():
                if entry.startswith("sqlalchemy.slaves."):
                    slave_path = entry.split(".")
                    slave_name = slave_path[2]
                    if slave_name == "master":
                        raise TGConfigError("A slave node cannot be named master")
                    slave_config = ".".join(slave_path[:3])
                    all_engines[slave_name] = slaves[slave_name] = engine_from_config(
                        conf, slave_config + "."
                    )

            if not conf["balanced_engines"]["slaves"]:
                raise TGConfigError(
                    "When running in balanced mode your must specify at least a slave node"
                )

        return engine



"""Utilities

The modules in the utilities package are expected to be independent
pure python modules that provide utility functions and classes that
can be used by any Python application.

They should not depend on TurboGears specific modules or objects,
in case they do, they should be moved to the support package.
"""

from .bunch import Bunch
from .decorators import no_warn
from .files import DottedFileLocatorError, DottedFileNameFinder  # noqa: F401
from .lazystring import LazyString, lazify

__all__ = (
    "Bunch",
    "no_warn",
    "DottedFileLocatorError",
    "DottedFileNameFinder",
    "LazyString",
    "lazify",
)



# -*- coding: utf-8 -*-
import datetime
import json
import re
from functools import partial
from json import loads

import pytest
from crank.util import get_params_with_argspec
from webtest import TestApp

import tests
import tg
from tests.base import (
    TestWSGIController,
    data_dir,
    make_app,
    setup_session_dir,
    teardown_session_dir,
)
from tg.configuration.utils import TGConfigError
from tg.configurator.fullstack import FullStackApplicationConfigurator
from tg.controllers import DecoratedController, TGController, abort
from tg.controllers.util import validation_errors_response
from tg.decorators import Decoration, before_call, before_render, expose, validate
from tg.i18n import lazy_ugettext as l_
from tg.support.converters import asbool, asint
from tg.validation import Convert, RequireValue, TGValidationError


def setup_module():
    setup_session_dir()

def teardown_module():
    teardown_session_dir()


class controller_based_validate(validate):
    def __init__(self, error_handler=None, *args, **kw):
        self.error_handler = error_handler
        self.needs_controller = True

        class Validators(object):
            def validate(self, controller, params):
                return params

        self.validators = Validators()

BoolValidator = Convert(asbool)
IntValidator = Convert(asint, msg="Please enter an integer value")

class EmailValidator:
    RE = re.compile(r"\w+@\w+\.\w+")

    @staticmethod
    def to_python(value):
        if not EmailValidator.RE.match(value):
            raise TGValidationError("Not an email address")
        return value

class ColonLessGenericValidator(object):
    def validate(self, value):
        raise TGValidationError('Unknown Error', value, {'_the_form':'Unknown Error'})

class ThrowAwayValidationIntentValidator(object):
    def validate(self, value):
        tg.request.validation.intent = None
        raise TGValidationError('Unknown Error', value, {'_the_form':'Unknown Error'})

def error_handler_function(controller_instance, uid, num):
    return 'UID: %s' % uid


def ControllerWrapperForErrorHandler(caller):
    def call(*args, **kw):
        value = caller(*args, **kw)
        return value + 'X'
    return call


class ErrorHandlerCallable(object):
    def __call__(self, controller_instance, uid, num):
        return 'UID: %s' % uid

class BasicTGController(TGController):
    @expose()
    @validate(ColonLessGenericValidator())
    def validator_without_columns(self, **kw):
        return tg.request.validation.errors['_the_form']

    @expose('json:')
    @validate(validators={"some_int": IntValidator})
    def validated_int(self, some_int):
        assert isinstance(some_int, int)
        return dict(response=some_int)

    @expose('json:')
    @validate(validators={"a": IntValidator})
    def validated_and_unvalidated(self, a, b):
        assert isinstance(a, int)
        assert isinstance(b, str)
        return dict(int=a, str=b)

    @expose()
    @controller_based_validate()
    def validate_controller_based_validator(self, *args, **kw):
        return 'ok'

    @expose('json:')
    @validate(validators={"a": IntValidator, "someemail": EmailValidator})
    def two_validators(self, a=None, someemail=None, *args):
        errors = tg.request.validation.errors
        values = tg.request.validation.values
        return dict(a=a, someemail=someemail,
                    errors=str(errors), values=str(values))

    @expose('json:')
    @validate(validators={"a": IntValidator})
    def with_default_shadow(self, a, b=None ):
        """A default value should not cause the validated value to disappear"""
        assert isinstance( a, int ), type(a)
        return {
            'int': a,
        }

    @expose('json:')
    @validate(validators={
        "a": IntValidator,"b":IntValidator,"c":IntValidator,"d":IntValidator
    })
    def with_default_shadow_long(self, a, b=None,c=None,d=None ):
        """A default value should not cause the validated value to disappear"""
        assert isinstance( a, int ), type(a)
        assert isinstance( b, int ), type(b)
        assert isinstance( c, int ), type(c)
        assert isinstance( d, int ), type(d)
        return {
            'int': [a,b,c,d],
        }

    @expose()
    def set_lang(self, lang=None):
        tg.session['tg_lang'] = lang
        tg.session.save()
        return 'ok'

    @expose('json:')
    @before_render(lambda rem,params,output:output.update({'GOT_ERROR':'HOOKED'}))
    def hooked_error_handler(self, *args, **kw):
        return dict(GOT_ERROR='MISSED HOOK')

    @expose()
    @validate({'v':IntValidator}, error_handler=hooked_error_handler)
    def with_hooked_error_handler(self, *args, **kw):
        return dict(GOT_ERROR='NO ERROR')

    @expose()
    def error_handler(self, *args, **kw):
        return 'ERROR HANDLER!'

    @expose('json:')
    @validate(validators={"some_int": IntValidator},
              error_handler=error_handler)
    def validate_other_error_handler(self, some_int):
        return dict(response=some_int)

    def unexposed_error_handler(self, uid, **kw):
        return 'UID: %s' % uid

    @expose()
    @validate({'uid': IntValidator,
               'num': IntValidator},
              error_handler=unexposed_error_handler)
    def validate_unexposed(self, uid, num):
        return 'HUH'

    @expose()
    @validate({'num': IntValidator},
              error_handler=partial(unexposed_error_handler,
                                    uid=5))
    def validate_partial(self, num):
        return 'HUH'

    @expose()
    @validate({'uid': IntValidator,
               'num': IntValidator},
              error_handler=error_handler_function)
    def validate_function(self, uid, num):
        return 'HUH'

    @expose()
    @validate({'uid': IntValidator,
               'num': IntValidator},
              error_handler=ErrorHandlerCallable())
    def validate_callable(self, uid, num):
        return 'HUH'

    @expose()
    @validate({'uid': IntValidator},
              error_handler=ErrorHandlerCallable())
    @validate({'num': IntValidator},
              error_handler=abort(412, error_handler=True))
    def validate_multi(self, uid, num):
        return str(uid+num)

    @expose()
    @validate({'uid': IntValidator},
              error_handler=abort(412, error_handler=True))
    def abort_error_handler(self):
        return 'HUH'

    @expose()
    @validate({'uid': IntValidator},
              error_handler=validation_errors_response)
    def validate_json_errors(self):
        return 'HUH'

    @expose()
    def validate_json_errors_complex_types(self, date):
        tg.request.validation.values = {'date': datetime.datetime.utcnow()}
        return validation_errors_response()

    @expose()
    @before_call(lambda remainder, params: params.setdefault('num', 5))
    def hooked_error_handler(self, uid, num):
        return 'UID: %s, NUM: %s' % (uid, num)

    @expose()
    @validate(ThrowAwayValidationIntentValidator(),
              error_handler=abort(412, error_handler=True))
    def throw_away_intent(self, uid):
        if tg.request.validation.exception:
            return 'ERROR'
        return 'UHU?'

    @expose()
    @validate(error_handler=hooked_error_handler)
    def passthrough_validation(self, uid):
        return str(uid)

    @expose()
    @validate({'uid': IntValidator},
              error_handler=hooked_error_handler)
    def validate_hooked(self, uid):
        return 'HUH'

    # Decorate validate_hooked with a controller wrapper
    Decoration.get_decoration(hooked_error_handler)\
        ._register_controller_wrapper(ControllerWrapperForErrorHandler)

    @expose()
    def manually_handle_validation(self):
        # This is done to check that we don't break compatibility
        # with external modules that perform custom validation like tgext.socketio

        controller = self.__class__.validate_function
        args = (2, 'NaN')
        context = tg.request.environ["tg.locals"]
        try:
            output = ''
            validate_params = get_params_with_argspec(controller, {}, args)
            params = DecoratedController._perform_validate(controller,
                                                           validate_params,
                                                           context)
        except TGValidationError as inv:
            obj, error_handler,_ = DecoratedController._process_validation_errors(
                controller, args, {}, inv, context
            )
            output = error_handler(obj, *args)
        return output

    @expose(content_type='text/plain')
    @validate({
        'num': Convert(int, l_('This must be a number'))
    }, error_handler=validation_errors_response)
    def post_pow2(self, num=-1):
        return str(num*num)

    @expose(content_type='text/plain')
    @validate({
        'num': Convert(int, l_('This must be a number'), default=0)
    }, error_handler=validation_errors_response)
    def post_pow2_opt(self, num=-1):
        return str(num*num)

    @expose(content_type='text/plain')
    @validate({
        'num': Convert(int, str('àèìòù'))
    }, error_handler=validation_errors_response)
    def unicode_error_pow(self, num=-1):
        return str(num*num)

    @expose(content_type='text/plain')
    @validate({
        'num': Convert(int, l_(str('àèìòù')))
    }, error_handler=validation_errors_response)
    def lazy_unicode_error_pow(self, num=-1):
        return str(num * num)

    @expose(content_type='text/plain')
    @validate({
        'val': Convert(lambda v: int(v) > 0 or int('ERROR'))
    }, error_handler=validation_errors_response)
    def chain_validation_0(self, val):
        return '>0'

    @expose(content_type='text/plain')
    @validate({
        'val': Convert(lambda v: int(v) > 1 or int('ERROR'))
    }, error_handler=chain_validation_0, chain_validation=True)
    def chain_validation_1(self, val):
        return '>1'

    @expose(content_type='text/plain')
    @validate({
        'val': Convert(lambda v: int(v) > 2 or int('ERROR'))
    }, error_handler=chain_validation_1, chain_validation=True)
    def chain_validation_2(self, val):
        return '>2'

    @expose(content_type='text/plain')
    @validate({
        'val': Convert(lambda v: int(v) > 3 or int('ERROR'))
    }, error_handler=chain_validation_2, chain_validation=True)
    def chain_validation_begin(self, val):
        return '>3'

    @expose(content_type='text/plain')
    @validate({
        'val': RequireValue(msg=l_("Value is required"))
    }, error_handler=validation_errors_response)
    def require_value(self, val=None):
        return val


class TestTGController(TestWSGIController):
    def setup_method(self):
        tg.config.update({
            'paths': {'root': data_dir},
            'package': tests,
        })

        self.app = make_app(BasicTGController, config_options={
            'i18n.enabled': True,
            'trace_errors.enable': False,
            'errorpage.enabled': False
        })
        TestWSGIController.setup_method(self)

    def test_basic_validation_and_jsonification(self):
        """Ensure you can pass in a dictionary of validators"""
        form_values = {"some_int": 22}
        resp = self.app.post('/validated_int', form_values)
        assert '{"response": 22}'in resp, resp

    def test_validation_other_error_handler(self):
        form_values = {"some_int": 'TEXT'}
        resp = self.app.post('/validate_other_error_handler', form_values)
        assert 'ERROR HANDLER!'in resp, resp

    def test_validator_without_columns(self):
        form_values = {"some_int": 22}
        resp = self.app.post('/validator_without_columns', form_values)
        assert 'Unknown Error' in resp, resp

    def test_for_other_params_after_validation(self):
        """Ensure that both validated and unvalidated data make it through"""
        form_values = {'a': 1, 'b': "string"}
        resp = self.app.post('/validated_and_unvalidated', form_values)
        assert '"int": 1' in resp
        assert '"str": "string"' in resp, resp

    def test_validation_shadowed_by_defaults( self ):
        """Catch regression on positional argument validation with defaults"""
        resp = self.app.post('/with_default_shadow/1?b=string')
        assert '"int": 1' in resp, resp

    def test_optional_shadowed_by_defaults( self ):
        """Catch regression on optional arguments being reverted to un-validated"""
        resp = self.app.post('/with_default_shadow_long/1?b=2&c=3&d=4')
        assert '"int": [1, 2, 3, 4]' in resp, resp

    def test_validation_fails_with_no_error_handler(self):
        form_values = {'a':'asdf', 'b':"string"}

        with pytest.raises(AssertionError):
            resp = self.app.post('/validated_and_unvalidated', form_values)

    def test_two_validators_errors(self):
        """Ensure that multiple validators are applied correctly"""
        form_values = {'a': '1', 'someemail': "guido@google.com"}
        resp = self.app.post('/two_validators', form_values)
        content = loads(resp.body.decode('utf-8'))
        assert content['a'] == 1

    def test_validation_errors(self):
        """Ensure that dict validation produces a full set of errors"""
        form_values = {'a': '1', 'someemail': "guido~google.com"}
        resp = self.app.post('/two_validators', form_values)
        content = loads(resp.body.decode('utf-8'))
        errors = content.get('errors', None)
        assert errors, 'There should have been at least one error'
        assert 'someemail' in errors, \
            'The email was invalid and should have been reported in the errors'

    def test_controller_based_validator(self):
        """Test controller based validation"""
        resp = self.app.post('/validate_controller_based_validator')
        assert 'ok' in resp, resp

    def test_hook_after_validation_error(self):
        resp = self.app.post('/with_hooked_error_handler?v=a')
        assert 'HOOKED' in resp, resp

    def test_validation_error_has_message(self):
        e = TGValidationError('This is a validation error')
        assert str(e) == 'This is a validation error'

    def test_validate_partial(self):
        resp = self.app.post('/validate_partial', {'num': 'NaN'})
        assert resp.text == 'UID: 5', resp

    def test_validate_unexposed(self):
        resp = self.app.post('/validate_unexposed', {'uid': 2,
                                                     'num': 'NaN'})
        assert resp.text == 'UID: 2', resp

    def test_validate_function(self):
        resp = self.app.post('/validate_function', {'uid': 2,
                                                    'num': 'NaN'})
        assert resp.text == 'UID: 2', resp

    def test_validate_callable(self):
        resp = self.app.post('/validate_callable', {'uid': 2,
                                                    'num': 'NaN'})
        assert resp.text == 'UID: 2', resp

    def test_validate_multi(self):
        resp = self.app.post('/validate_multi', {'uid': 'NaN', 'num': 2})
        assert resp.text == 'UID: NaN', resp

        resp = self.app.post('/validate_multi', {'uid': 2, 'num': 'NaN'}, status=412)
        assert resp.status.startswith('412')

        resp = self.app.post('/validate_multi', {'uid': 2, 'num': 2})
        assert resp.text == '4', resp

        resp = self.app.post('/validate_multi', {'uid': 'NaN', 'num': 'NaN'})
        assert resp.text == 'UID: NaN', resp

    def test_validate_hooked(self):
        resp = self.app.post('/validate_hooked', {'uid': 'NaN'})
        assert resp.text == 'UID: NaN, NUM: 5X', resp

    def test_manually_handle_validation(self):
        # This is done to check that we don't break compatibility
        # with external modules that perform custom validation like tgext.socketio
        resp = self.app.post('/manually_handle_validation')
        assert resp.text == 'UID: 2', resp

    def test_abort_error_handler(self):
        resp = self.app.post('/abort_error_handler', {'uid': 'NaN'}, status=412)
        assert resp.status.startswith('412')

    def test_json_error_handler(self):
        resp = self.app.post('/validate_json_errors', {'uid': 'NaN'}, status=412)
        assert resp.json['errors']['uid'] == 'Please enter an integer value'

    def test_json_error_handler_complex_type(self):
        resp = self.app.post('/validate_json_errors_complex_types', {'date': '2014-01-01'},
                             status=412)
        assert resp.json['values']['date'] == '2014-01-01', resp

    def test_throw_way_validation_intent(self):
        # This should actually never happen.
        # It requires the validator to mess up with TG internals, it's just to provide full coverage.
        resp = self.app.post('/throw_away_intent', {'uid': 5})
        assert resp.text == 'ERROR', resp

    def test_passthrough_validation(self):
        # This should actually never happen.
        # It requires the validator to mess up with TG internals, it's just to provide full coverage.
        resp = self.app.post('/passthrough_validation', {'uid': 5})
        assert resp.text == '5', resp

    def test_convert_validation(self):
        resp = self.app.post('/post_pow2', {'num': '5'})
        assert resp.text == '25', resp

    def test_convert_validation_fail(self):
        resp = self.app.post('/post_pow2', {'num': 'HELLO'}, status=412)
        assert 'This must be a number' in resp.json['errors']['num']

    def test_convert_validation_missing(self):
        resp = self.app.post('/post_pow2', {'num': ''}, status=412)
        assert 'This must be a number' in resp.json['errors']['num']

        resp = self.app.post('/post_pow2', status=412)
        assert 'This must be a number' in resp.json['errors']['num']

    def test_convert_validation_optional(self):
        resp = self.app.post('/post_pow2_opt', {'num': 'HELLO'}, status=412)
        assert 'This must be a number' in resp.json['errors']['num']

        resp = self.app.post('/post_pow2_opt', {'num': '5'})
        assert resp.text == '25', resp

        resp = self.app.post('/post_pow2_opt', {'num': ''})
        assert resp.text == '0', resp

        resp = self.app.post('/post_pow2_opt')
        assert resp.text == '0', resp

    def test_validation_errors_unicode(self):
        resp = self.app.post('/unicode_error_pow', {'num': 'NOT_A_NUMBER'}, status=412)
        assert resp.json['errors']['num'] == str('àèìòù'), resp.json

    def test_validation_errors_lazy_unicode(self):
        resp = self.app.post('/lazy_unicode_error_pow', {'num': 'NOT_A_NUMBER'}, status=412)
        assert resp.json['errors']['num'] == str('àèìòù'), resp.json

    def test_requirevalue_validation(self):
        resp = self.app.post('/require_value', {"val": "hello"})
        assert resp.text == 'hello', resp

        resp = self.app.post('/require_value', {}, status=412)
        assert resp.json["errors"]["val"] == 'Value is required', resp


class TestChainValidation(TestWSGIController):
    def setup_method(self):
        tg.config.update({
            'paths': {'root': data_dir},
            'package': tests,
        })

        self.app = make_app(BasicTGController, config_options={
            'i18n.enabled': True
        })

        TestWSGIController.setup_method(self)

    def test_no_chain_validation(self):
        res = self.app.get('/chain_validation_begin', params={'val': 4})
        assert res.text == '>3'

        res = self.app.get('/chain_validation_begin', params={'val': 3})
        assert res.text == '>2'

    def test_single_chain_validation(self):
        res = self.app.get('/chain_validation_begin', params={'val': 2})
        assert res.text == '>1'

    def test_double_chain_validation(self):
        res = self.app.get('/chain_validation_begin', params={'val': 1})
        assert res.text == '>0'

    def test_last_chain_validation(self):
        res = self.app.get('/chain_validation_begin', params={'val': 0}, status=412)
        assert res.json == json.loads('{"errors":{"val":"Invalid"},"values":{"val":"0"}}')


class TestValidationConfiguration:
    def create_app(self, root_controller, options=None):
        cfg = FullStackApplicationConfigurator()
        cfg.update_blueprint({'root_controller': root_controller})
        cfg.update_blueprint(options or {})

        app = TestApp(cfg.make_wsgi_app({
            'debug': False,
            'errorpage.enabled': False,
            'trace_errors.enable': False
        }, {}))
        return app

    def test_no_validation_function(self):
        class FakeSchema:
            pass

        class RootController(TGController):
            @validate(FakeSchema())
            @expose("text/plain")
            def test(self, **kwargs):
                return "HI"

        app = self.create_app(RootController())

        with pytest.raises(TGConfigError) as exc_info:
            app.get("/test", {"value": 5})
        assert "No validation validator function found for" in str(exc_info.value)
        assert "FakeSchema" in str(exc_info.value)

    def test_custom_validation(self):
        class FakeSchema:
            pass

        def validate_fake_schema(schema, params):
            if params.get("fail"):
                raise TGValidationError("Invalid params", value=params,
                                        error_dict={"fail": "Fail is true"})
            return params

        class RootController(TGController):
            @validate(FakeSchema())
            @expose("text/plain")
            def test(self, **kwargs):
                if tg.request.validation.errors:
                    return str(tg.request.validation.errors)
                return "HI"

        app = self.create_app(RootController(), {
            "validation.validators": {FakeSchema: validate_fake_schema}
        })

        resp = app.get("/test", {"value": 5})
        assert resp.text == "HI"

        resp = app.get("/test", {"fail": 1})
        assert "Fail is true" in resp.text

    def test_custom_validation_error(self):
        class FakeSchema:
            pass

        def validate_fake_schema(schema, params):
            if params.get("fail"):
                raise FakeError()
            return params

        class FakeError(Exception):
            pass

        def explode_fake_error(error):
            return {"values": {"fail": True}, "errors": "fail was true"}

        class RootController(TGController):
            @validate(FakeSchema())
            @expose("text/plain")
            def test(self, **kwargs):
                if tg.request.validation.errors:
                    return str(tg.request.validation.errors)
                return "HI"

        app = self.create_app(RootController(), {
            "validation.validators": {FakeSchema: validate_fake_schema},
            "validation.exceptions": [FakeError],
            "validation.explode": {FakeError: explode_fake_error}
        })

        resp = app.get("/test", {"value": 5})
        assert resp.text == "HI"

        resp = app.get("/test", {"fail": 1})
        assert "fail was true" in resp.text

        # Do not provide explode function
        app = self.create_app(RootController(), {
            "validation.validators": {FakeSchema: validate_fake_schema},
            "validation.exceptions": [FakeError]
        })
        with pytest.raises(TGConfigError) as exc_info:
            resp = app.get("/test", {"fail": 1})
        assert "No validation explode function found for" in str(exc_info.value)
        assert "FakeError" in str(exc_info.value)

        # Do not provide missing explode function
        app = self.create_app(RootController(), {
            "validation.validators": {FakeSchema: validate_fake_schema},
            "validation.exceptions": [FakeError],
            "validation.explode": {FakeError: None}
        })
        with pytest.raises(TGConfigError) as exc_info:
            resp = app.get("/test", {"fail": 1})
        assert "No validation explode function found for" in str(exc_info.value)
        assert "FakeError" in str(exc_info.value)



def im_func(f):
    return getattr(f, "__func__", None)


def default_im_func(f):
    return getattr(f, "__func__", f)


def im_self(f):
    return getattr(f, "__self__", None)


def im_class(f):
    self = im_self(f)
    if self is not None:
        return self.__class__
    else:
        return None
